# -*- coding: utf-8 -*-
"""
Created on Fri Jul 15 16:50:05 2022

@author: felix
"""

from ..Utility_Functions import file_io as fio
import numpy as np 


# is there a quick way to compute the occupied bbox density. 
def remove_very_large_bbox(boxes, shape, thresh=0.5, aspect_ratio=None, method='density', max_density=4, mode='fast', return_ind=False):
    
    from sklearn.metrics.pairwise import pairwise_distances
    
    if method == 'density':
        
        boxes_ = boxes.copy()
        keep_ind = np.arange(len(boxes_))
        
        if aspect_ratio is not None:
            w = boxes[:,2] - boxes[:,0]
            h = boxes[:,3] - boxes[:,1]
            wh = np.vstack([w,h])
            
            aspect_w_h = np.max(wh, axis=0) / (np.min(wh, axis=0) + .1)
            boxes_ = boxes_[aspect_w_h<=aspect_ratio]
            keep_ind = keep_ind[aspect_w_h<=aspect_ratio]
        
        box_density = np.zeros(shape)
        bbox_coverage = np.zeros(len(boxes_))
        box_centroids_x = np.clip((.5*(boxes_[:,0] + boxes_[:,2])).astype(np.int32), 0, shape[1]-1).astype(np.int32)
        box_centroids_y = np.clip((.5*(boxes_[:,1] + boxes_[:,3])).astype(np.int32), 0, shape[0]-1).astype(np.int32)
        
        if mode == 'fast':
            box_centroids = np.vstack([box_centroids_x, box_centroids_y]).T
            box_centroids_r = np.sqrt((boxes_[:,2]-boxes_[:,0])*(boxes_[:,3]-boxes_[:,1])/np.pi +.1)
        
            box_centroids_distance = pairwise_distances(box_centroids)
            bbox_coverage = np.nansum(box_centroids_distance<=box_centroids_r[:,None], axis=1)
            
        else:
            # box_density[box_centroids_y, box_centroids_x] += 1
            for cent_ii in np.arange(len(box_centroids_x)):
                cent_x = box_centroids_x[cent_ii]
                cent_y = box_centroids_y[cent_ii]
                box_density[int(cent_y), int(cent_x)] += 1
            
            for box_ii, box in enumerate(boxes_):
                x1, y1, x2, y2 = box
                x1 = np.clip(int(x1), 0, shape[1]-1)
                y1 = np.clip(int(y1), 0, shape[0]-1)
                x2 = np.clip(int(x2), 0, shape[1]-1)
                y2 = np.clip(int(y2), 0, shape[0]-1)
                bbox_coverage[box_ii] = np.nansum(box_density[int(y1):int(y2), 
                                                              int(x1):int(x2)])        
        # print(bbox_coverage)
        if return_ind == False:
            return boxes_[bbox_coverage<=max_density]
        else:
            return boxes_[bbox_coverage<=max_density], keep_ind[bbox_coverage<=max_density]

    if method == 'area':
        areas_box = []
        area_shape = float(np.prod(shape))
        
    #    print(area_shape)
        for box in boxes:
            x1, y1, x2, y2 = box
            x1 = int(x1)
            y1 = int(y1)
            x2 = int(x2)
            y2 = int(y2)
            
            area_box = (y2-y1)*(x2-x1)
            areas_box.append(area_box)
            
        areas_box = np.hstack(areas_box)
        areas_box_frac = areas_box / float(area_shape)
        return boxes[areas_box_frac<=thresh]



def Eval_dense_optic_flow(prev, present, params):
    r""" Computes the optical flow using Farnebacks Method

    Parameters
    ----------
    prev : numpy array
        previous frame, m x n image
    present :  numpy array
        current frame, m x n image
    params : Python dict
        a dict object to pass all algorithm parameters. Fields are the same as that in the opencv documentation, https://docs.opencv.org/2.4/modules/video/doc/motion_analysis_and_object_tracking.html. Our recommended starting values:
                
            * params['pyr_scale'] = 0.5
            * params['levels'] = 3
            * params['winsize'] = 15
            * params['iterations'] = 3
            * params['poly_n'] = 5
            * params['poly_sigma'] = 1.2
            * params['flags'] = 0
        
    Returns
    -------
    flow : finds the displacement field between frames, prev and present such that :math:`\mathrm{prev}(y,x) = \mathrm{next}(y+\mathrm{flow}(y,x)[1], x+\mathrm{flow}(y,x)[0])` where (x,y) is the cartesian coordinates of the image.
    """
    
    import numpy as np 
    import warnings
    import cv2

    # Check version of opencv installed, if not 3.0.0 then issue alert.
#    if '3.0.0' in cv2.__version__ or '3.1.0' in cv2.__version__:
        # Make the image pixels into floats.
    prev = prev.astype(np.float32)
    present = present.astype(np.float32)

    if cv2.__version__.split('.')[0] == '3' or cv2.__version__.split('.')[0] == '4':
        flow = cv2.calcOpticalFlowFarneback(prev, present, None, params['pyr_scale'], params['levels'], params['winsize'], params['iterations'], params['poly_n'], params['poly_sigma'], params['flags']) 
    if cv2.__version__.split('.')[0] == '2':
        flow = cv2.calcOpticalFlowFarneback(prev, present, pyr_scale=params['pyr_scale'], levels=params['levels'], winsize=params['winsize'], iterations=params['iterations'], poly_n=params['poly_n'], poly_sigma=params['poly_sigma'], flags=params['flags']) 
#    print(flow.shape)
    return flow


def rescale_intensity_percent(img, intensity_range=[2,98]):

    from skimage.exposure import rescale_intensity
    import numpy as np 

    p2, p98 = np.percentile(img, intensity_range)
    img_ = rescale_intensity(img, in_range=(p2,p98))

    return img_


# add in optical flow. 
def extract_optflow(vid, optical_flow_params, rescale_intensity=True, intensity_range=[2,98]): 
    # uses CV2 built in farneback ver. which is very fast and good for very noisy and small motion
    import cv2
    from skimage.exposure import rescale_intensity
    import numpy as np
    from tqdm import tqdm 

    vid_flow = []
    n_frames = len(vid)

    for frame in tqdm(np.arange(len(vid)-1)):
        frame0 = rescale_intensity_percent(vid[frame], intensity_range=intensity_range)
        frame1 = rescale_intensity_percent(vid[frame+1], intensity_range=intensity_range)
        flow01 = Eval_dense_optic_flow(frame0, frame1, 
                                       params=optical_flow_params)
        vid_flow.append(flow01)
    vid_flow = np.array(vid_flow).astype(np.float32) # to save some space. 

    return vid_flow



def predict_new_boxes_flow_tf(boxes, flow):
    
    from skimage.transform import estimate_transform, matrix_transform, SimilarityTransform
    from skimage.measure import ransac
    import numpy as np 
    
    flow_x = flow[:,:,0]
    flow_y = flow[:,:,1]
      
    new_boxes_tf = []
    tfs = []
    
    for box in boxes:
        x1,y1,x2,y2 = box
        
        x1 = int(x1); y1 =int(y1); x2=int(x2); y2 =int(y2);  # added, weirdly ...
        
        # how to take into account the change in size (scale + translation ? very difficult. )
        flow_x_patch = flow_x[y1:y2, x1:x2].copy()
        flow_y_patch = flow_y[y1:y2, x1:x2].copy()
        
        nrows_, ncols_ = flow_x_patch.shape
        
        # threshold_the mag
        mag_patch = np.sqrt(flow_x_patch ** 2 + flow_y_patch ** 2)
        select = mag_patch > 0
#        select = np.ones(mag_patch.shape) > 0
        pix_X, pix_Y = np.meshgrid(range(ncols_), range(nrows_))
        
        if np.sum(select) == 0:
            # if we cannot record movement in the box just append the original ?
            tfs.append([])
            new_boxes_tf.append([x1,y1,x2,y2])
        else:
            src_x = pix_X[select].ravel(); dst_x = src_x + flow_x_patch[select]
            src_y = pix_Y[select].ravel(); dst_y = src_y + flow_y_patch[select]
            src = np.hstack([src_x[:,None], src_y[:,None]])
            dst = np.hstack([dst_x[:,None], dst_y[:,None]])
            
            # estimate the transformation. 
            matrix = estimate_transform('similarity', src[:,[0,1]], dst[:,[0,1]])
            tf_scale = matrix.scale; tf_offset = matrix.translation
            
#            print tf_scale, tf_offset
            if np.isnan(tf_scale):
                tfs.append(([]))
                new_boxes_tf.append([x1,y1,x2,y2])
            else:
                x = .5*(x1+x2); w = x2-x1
                y = .5*(y1+y2); h = y2-y1
                
                x1_tf_new = x + tf_offset[0] - w*tf_scale/2.
                y1_tf_new = y + tf_offset[1] - h*tf_scale/2.
                x2_tf_new = x1_tf_new + w*tf_scale
                y2_tf_new = y1_tf_new + h*tf_scale
                
        #        print x1_tf_new
                tfs.append(matrix)
                new_boxes_tf.append([x1_tf_new, y1_tf_new, x2_tf_new, y2_tf_new])
        
    new_boxes_tf = np.array(new_boxes_tf)
    
    return tfs, new_boxes_tf

# this version is just over the boxes. given voc format. 
def bbox_iou_corner_xy(bboxes1, bboxes2):
    
    import numpy as np 
    
    """
    computes the distance matrix between two sets of bounding boxes.
    Args:
        bboxes1: shape (total_bboxes1, 4)
            with x1, y1, x2, y2 point order.
        bboxes2: shape (total_bboxes2, 4)
            with x1, y1, x2, y2 point order.

        p1 *-----
           |     |
           |_____* p2

    Returns:
        Tensor with shape (total_bboxes1, total_bboxes2)
        with the IoU (intersection over union) of bboxes1[i] and bboxes2[j]
        in [i, j].
    """

    x11, y11, x12, y12 = bboxes1[:,0], bboxes1[:,1], bboxes1[:,2], bboxes1[:,3]
    x21, y21, x22, y22 = bboxes2[:,0], bboxes2[:,1], bboxes2[:,2], bboxes2[:,3]


    x11 = x11[:,None]; y11 = y11[:,None]; x12=x12[:,None]; y12=y12[:,None]
    x21 = x21[:,None]; y21 = y21[:,None]; x22=x22[:,None]; y22=y22[:,None]

    xI1 = np.maximum(x11, np.transpose(x21))
    xI2 = np.minimum(x12, np.transpose(x22))

    yI1 = np.maximum(y11, np.transpose(y21))
    yI2 = np.minimum(y12, np.transpose(y22))

    inter_area = np.maximum((xI2 - xI1), 0.) * np.maximum((yI2 - yI1), 0.)

    bboxes1_area = (x12 - x11) * (y12 - y11)
    bboxes2_area = (x22 - x21) * (y22 - y21)

    union = (bboxes1_area + np.transpose(bboxes2_area)) - inter_area

    # some invalid boxes should have iou of 0 instead of NaN
    # If inter_area is 0, then this result will be 0; if inter_area is
    # not 0, then union is not too, therefore adding a epsilon is OK.
    return inter_area / (union+0.0001)


def bbox_tracks_2_array(vid_bbox_tracks, nframes, ndim):

    import numpy as np 
    N_tracks = len(vid_bbox_tracks)

    vid_bbox_tracks_all_array = np.zeros((N_tracks, nframes, ndim))
    vid_bbox_tracks_all_array[:] = np.nan

    vid_bbox_tracks_prob_all_array = np.zeros((N_tracks, nframes))
    vid_bbox_tracks_prob_all_array[:] = np.nan

    for ii in np.arange(N_tracks):
        tra_ii = vid_bbox_tracks[ii]
        tra_ii_times = np.array([tra_iii[0] for tra_iii in tra_ii])
        tra_ii_boxes = np.array([tra_iii[-1] for tra_iii in tra_ii]) # boxes is the last array. 
        tra_ii_prob = np.array([tra_iii[1] for tra_iii in tra_ii])

        vid_bbox_tracks_all_array[ii,tra_ii_times] = tra_ii_boxes.copy()
        vid_bbox_tracks_prob_all_array[ii,tra_ii_times] = tra_ii_prob.copy()

    return vid_bbox_tracks_prob_all_array, vid_bbox_tracks_all_array


def bbox_scalar_2_array(vid_scalar_tracks, nframes):

    import numpy as np 
    N_tracks = len(vid_scalar_tracks)

    vid_scalar_tracks_all_array = np.zeros((N_tracks, nframes))
    vid_scalar_tracks_all_array[:] = np.nan

    for ii in np.arange(N_tracks):
        tra_ii = vid_scalar_tracks[ii]
        tra_ii_times = np.array([tra_iii[0] for tra_iii in tra_ii])
        tra_ii_vals = np.array([tra_iii[1] for tra_iii in tra_ii])

        vid_scalar_tracks_all_array[ii,tra_ii_times] = tra_ii_vals.copy()

    return vid_scalar_tracks_all_array



def unpad_bbox_tracks(all_uniq_tracks, pad, shape, nframes, aspect_ratio=3): 
    
    import numpy as np 
    # define the bounds. 
    bounds_x = [0, shape[1]-1]
    bounds_y = [0, shape[0]-1] 
    
    all_uniq_bbox_tracks_new = []
    
    for tra_ii in np.arange(len(all_uniq_tracks)):

        all_uniq_tracks_ii = all_uniq_tracks[tra_ii]
        # all_track_labels_ii = track_labels[tra_ii]
        all_uniq_bbox_track_ii = []

        for jj in np.arange(len(all_uniq_tracks_ii)):
            
            tra = all_uniq_tracks_ii[jj]
            # tra_label = all_track_labels_ii[jj][-1]
            tp, conf, bbox = tra
            x1, y1, x2, y2 = bbox
            
            x1 = np.clip(x1, 0, shape[1]-1+2*pad)
            x2 = np.clip(x2, 0, shape[1]-1+2*pad)
            y1 = np.clip(y1, 0, shape[0]-1+2*pad)
            y2 = np.clip(y2, 0, shape[0]-1+2*pad)
    
            x1_= x1-pad 
            y1_= y1-pad 
            x2_ = x2-pad
            y2_ = y2-pad
            # print(x1_, x2_, y1_, y2_)
            
            """
            This is more complex ... 
                if y < 0 or  y > shape[0], then we have to flip x coordinate and flip y-coordinate.  
                # test this inversion ! ######## ---> for a fixed timepoint, just to double ceck we are getting the correct alogrithm. !. 
            """
            # left top 
            if y1_ < bounds_y[0] and (x1_ < bounds_x[0]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y1_ = -1 * y1_  # flip y 
                # x1_ = (bounds_x[1]+1*pad + x1_) - bounds_x[1] # flip x axis. 
                x1_ = bounds_x[0] - x1_ 
    
            if y2_ < bounds_y[0] and (x2_ < bounds_x[0]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y2_ = -1 * y2_ 
                # x2_ = (bounds_x[1]+1*pad + x2_) - bounds_x[1] # flip x axis. 
                x2_ = bounds_x[0] - x2_ 
            # middle top # this is ok 
            if y1_ < bounds_y[0] and (x1_ >= bounds_x[0] and x1_ <= bounds_x[1]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y1_ = -1 * y1_ 
                x1_ = bounds_x[1] - x1_ # flip x axis. 
    
            if y2_ < bounds_y[0] and (x2_ >= bounds_x[0] and x2_ <= bounds_x[1]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y2_ = -1 * y2_ 
                x2_ = bounds_x[1] - x2_ # flip x axis. 
                
            # right top 
            if y1_ < bounds_y[0] and (x1_ > bounds_x[1]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y1_ = -1 * y1_  # flip y 
                x1_ = bounds_x[1] - (x1_ - bounds_x[1])
    
            if y2_ < bounds_y[0] and (x2_ > bounds_x[1]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y2_ = -1 * y2_ 
                x2_ = bounds_x[1] - (x2_ - bounds_x[1])
    
            # middle left 
            if (y1_ >= bounds_y[0] and y1_ <= bounds_y[1]) and (x1_ < bounds_x[0]):
                # simply shift the x1_ back 
                x1_ = x1_ + (bounds_x[1]+1)
            if (y2_ >= bounds_y[0] and y2_ <= bounds_y[1]) and (x2_ < bounds_x[0]): 
                x2_ = x2_ + (bounds_x[1]+1)
    
            # middle right
            if (y1_ >= bounds_y[0] and y1_ <= bounds_y[1]) and (x1_ > bounds_x[1]):
                # simply shift the x1_ back 
                x1_ = x1_ - (bounds_x[1]+1)
            if (y2_ >= bounds_y[0] and y2_ <= bounds_y[1]) and (x2_ > bounds_x[1]):
                # simply shift the x1_ back 
                x2_ = x2_ - (bounds_x[1]+1)
    
            # left bottom 
            if y1_ > bounds_y[1] and (x1_ < bounds_x[0]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y1_ = bounds_y[1] + (bounds_y[1] - y1_) # flipping at the other end.  
                # x1_ = (bounds_x[1]+1*p-d + x1_) - bounds_x[1]  # flip x axis.
                x1_ = bounds_x[0] - x1_ 
            if y2_ > bounds_y[1] and (x2_ < bounds_x[0]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y2_ = bounds_y[1] + (bounds_y[1] - y2_) # flipping at the other end. 
                x2_ = bounds_x[0] - x2_ 
                # x2_ = (bounds_x[1]+1*pad + x2_) - bounds_x[1]
                # x2_ = x2_ - (bounds_x[1]+1)
                
            # middle bottom 
            if y1_ > bounds_y[1] and (x1_ >= bounds_x[0] and x1_ <= bounds_x[1]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y1_ = bounds_y[1] + (bounds_y[1] - y1_) # flipping at the other end.  
                x1_ = bounds_x[1] - x1_ # flip x axis. 
    
            if y2_ > bounds_y[1] and (x2_ >= bounds_x[0] and x2_ <= bounds_x[1]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y2_ = bounds_y[1] + (bounds_y[1] - y2_) # flipping at the other end. 
                x2_ = bounds_x[1] - x2_ # flip x axis. 
    
            # right bottom  
            if y1_ > bounds_y[1] and (x1_ > bounds_x[1]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y1_ = bounds_y[1] + (bounds_y[1] - y1_) # flipping at the other end.  
                # x1_ = x1_ - (bounds_x[1]+1*pad) + bounds_x[1] # flip x axis. 
                x1_ = bounds_x[1] - (x1_ - bounds_x[1])
            if y2_ > bounds_y[1] and (x2_ > bounds_x[1]):
                # then we flip the x axis and flip the y coordinate into the shape. (=180 rotation) 
                y2_ = bounds_y[1] + (bounds_y[1] - y2_) # flipping at the other end. 
                # x2_ = x2_ - (bounds_x[1]+1*pad) + bounds_x[1] # flip x axis. 
                x2_ = bounds_x[1] - (x2_ - bounds_x[1])
    
            """
            need a accuracy test for the designated label if the label is not the background label!. 
            """
            # we check the sign change here! to know if we need to modify.  
            w1_ = (x2_ - x1_) # x2 must be higher? if negative that means x1 needs to be shifted back. 
            h1_ = (y2_ - y1_) # if negative means y1 needs to be shifted 
            # print(w1_, h1_, (x1_,y1_,x2_,y2_))
            if w1_ < 0:
                # we have to shift but dunno which way. 
                x1_test = np.clip(x1_ - (bounds_x[1]), 0, bounds_x[1])
                x2_test = np.clip(x2_ + (bounds_x[1]), 0, bounds_x[1])
    
                area_x1_test = np.abs(x2_ - x1_test) * np.abs(y2_-y1_)
                area_x2_test = np.abs(x2_test - x1_) * np.abs(y2_-y1_)
                
                if area_x1_test > area_x2_test:
                    x1_ = x1_test
                else:
                    x2_ = x2_test
    
            if h1_ < 0:
                # y1_ = np.clip(y1_ - (bounds_y[1]+1), 0, bounds_y[1])
                # y2_ = np.clip(y2_ + (bounds_y[1]+1), 0, bounds_y[1])
                # we have to shift but dunno which way. 
                y1_test = np.clip(y1_ - (bounds_y[1]), 0, bounds_y[1]) # make y1 smaller or 
                y2_test = np.clip(y2_ + (bounds_y[1]), 0, bounds_y[1]) # make y2 bigger
    
                area_y1_test = np.abs(y2_ - y1_test) * np.abs(x2_ - x1_)
                area_y2_test = np.abs(y2_test - y1_) * np.abs(x2_ - x1_)
                
                if area_y1_test > area_y2_test:
                    y1_ = y1_test
                else:
                    y2_ = y2_test
            
            bbox_new = np.hstack([x1_, y1_, x2_, y2_])
            all_uniq_bbox_track_ii.append([tp, conf, bbox_new])
        all_uniq_bbox_tracks_new.append(all_uniq_bbox_track_ii)

        # convert this !. 
        # compile into a numpy array and send this out too 
        _, all_uniq_bbox_tracks_array_new = bbox_tracks_2_array(all_uniq_bbox_tracks_new, nframes=nframes, ndim=4)
        all_uniq_bbox_tracks_centroids_xy_new = np.array([.5*(all_uniq_bbox_tracks_array_new[...,0] + all_uniq_bbox_tracks_array_new[...,2]), 
                                                          .5*(all_uniq_bbox_tracks_array_new[...,1] + all_uniq_bbox_tracks_array_new[...,3])])
        
    return all_uniq_bbox_tracks_new, all_uniq_bbox_tracks_array_new, all_uniq_bbox_tracks_centroids_xy_new


def assign_label_detection_to_track_frame_by_frame(bbox_array_time, bbox_detections_time, min_iou=.25):

    """
    bbox_array_time: N_tracks x N_frames x 4
    bbox_detections_time: N_frames list 
    """
    import numpy as np 
    from scipy.optimize import linear_sum_assignment

    n_tracks, n_frames, _ = bbox_array_time.shape
    bbox_array_time_labels = np.zeros((n_tracks, n_frames), dtype=np.float64) # initialise to background!. 
    bbox_array_time_labels[:] = np.nan # initialise to empty 

    # iterate over time 
    for tt in np.arange(n_frames)[:]:
        # reference
        bbox_frame_tt = bbox_detections_time[tt].copy()
        bbox_labels_tt = bbox_frame_tt[:,0].copy()
        bbox_labels_bbox = bbox_frame_tt[:,1:5].copy()

        tracks_bboxes = bbox_array_time[:,tt].copy()
        non_nan_select = np.logical_not(np.isnan(tracks_bboxes[:,0])) # just test the one coordinate. # only these need to be given a label!. 

        tracks_bboxes_bbox = tracks_bboxes[non_nan_select>0].copy()
        labels_non_nan_select = np.zeros(len(tracks_bboxes_bbox), dtype=np.int32) # preinitialise 

        # build the iou cost matrix. between rows: tracks and cols: the frame boxes. 
        iou_matrix = bbox_iou_corner_xy(tracks_bboxes_bbox, 
                                        bbox_labels_bbox) 
        iou_matrix = np.clip(1.-iou_matrix, 0, 1) # to make it a dissimilarity matrix. 
        # solve the pairing problem.
        ind_ii, ind_jj = linear_sum_assignment(iou_matrix)
        
        # threshold as the matching is maximal. 
        iou_ii_jj = iou_matrix[ind_ii, ind_jj].copy()
        keep = iou_ii_jj <= (1-min_iou) # keep those with a minimal distance. 

        # keep = iou_ii_jj <= dist_thresh
        ind_ii = ind_ii[keep>0]; 
        ind_jj = ind_jj[keep>0]
        
        labels_non_nan_select[ind_ii] = bbox_labels_tt[ind_jj].copy()

        # copy back into the labels. 
        bbox_array_time_labels[non_nan_select>0, tt] = labels_non_nan_select.copy() # copy these assignments in now. 

    return bbox_array_time_labels


def compute_labeled_to_unlabeled_track(track_label_array):

    import numpy as np 

    n_tracks, n_frames = track_label_array.shape

    track_counts = []
    for tra_ii in np.arange(n_tracks):
        tra_label = track_label_array[tra_ii].copy()
        valid = np.logical_not(np.isnan(tra_label))

        num_valid = np.sum(valid)
        valid_labels = tra_label[valid].astype(np.int32)

        num_nonzeros = np.sum(valid_labels > 0) 
        track_counts.append([num_valid, num_nonzeros, float(num_nonzeros) / num_valid, n_frames])
    track_counts = np.vstack(track_counts)

    return track_counts


def moving_average_bbox_tracks(list_tracks, avg_func=np.nanmean, winsize=3, min_winsize_prop=0.1, pad_mode='edge', *args):

    import numpy as np 

    list_tracks_smooth = []

    for tra_ii in np.arange(len(list_tracks)):

        track_ii = np.array(list_tracks[tra_ii]) # numpy array 
        winsize_track = winsize
        if winsize/float(len(track_ii)) < min_winsize_prop: 
            # then reduce the winsize smoothing!.   
            winsize_track = np.maximum(3, int(len(track_ii)*min_winsize_prop))
        track_ii_pad = np.pad(track_ii, [[winsize_track//2, winsize_track//2], [0,0]], mode=pad_mode, *args)

        track_ii_smooth = np.vstack([avg_func(track_ii_pad[tt:tt+winsize_track], axis=0) for tt in np.arange(len(track_ii))])
        list_tracks_smooth.append(track_ii_smooth)

    return list_tracks_smooth


# using optical flow to implement a predictive tracker for blebs. 
def track_bleb_bbox(vid_flow, # flow is already computed. 
                    vid_bboxes, # bbox files for each frame only !.   
                    vid=None, # only for visualisation purposes. 
                    iou_match=.25,
                    ds_factor = 1,
                    wait_time=10,
                    min_aspect_ratio=3,
                    max_dense_bbox_cover=8,
                    min_tra_len_filter=5,
                    min_tra_lifetime_ratio = .1, 
                    to_viz=True,
                    remove_eccentric_bbox=False,
                    saveanalysisfolder=None):

    """
    uses optical flow and registration as a predictor to improve tracking over occlusion.
    # we should preserve the bbox coordinates + the given probability from yolo.....  that way we can always do further processing? 
    """
    import numpy as np 
    import seaborn as sns
    import pylab as plt 
    import os 
    from tqdm import tqdm 
    from scipy.optimize import linear_sum_assignment

    """
    initial setting up. 
    """
    im_shape = vid_flow[0].shape[:2]
    nframes = len(vid_flow)+1 # the number of frames in the video.

    if saveanalysisfolder is not None:
        print('saving graphics in folder: ', saveanalysisfolder)
        # saveanalysisfolder_movie = os.path.join(saveanalysisfolder, 'track_boundaries'); mkdir(saveanalysisfolder_movie); 
        saveanalysisfolder_movie_bbox = os.path.join(saveanalysisfolder, 'track_bbox'); 
        mkdir(saveanalysisfolder_movie_bbox);     

    # =============================================================================
    #     BBox Tracks
    # =============================================================================

    # terminated_vid_cell_tracks = [] 
    terminated_vid_bbox_tracks = []
    terminated_check_match_bbox_tracks = []

    vid_bbox_tracks = [] # this is to keep track of the actual bbox we use, including inferred. 
    vid_bbox_tracks_last_time = [] # record when the last time a proper match was made. 
    vid_match_check_bbox_tracks = []

# =============================================================================
#     build up the tracks dynamically now, frame by frame dynamically , with a track waiting time before termination. 
# ============================================================================
    for ii in tqdm(np.arange(nframes-1)[:]):
        """
        if working set is empty or otherwise, then add to working set. 
        """
        # add these to the working tracks. 
        if ii == 0 or len(vid_bbox_tracks)==0: 
            # bboxfile_ii = vid_bboxes[ii]
            """
            swap this element out. 
            """
            # prob_ii, boxes_ii = load_bbox_frame_voc( vid[0], bboxfile_ii) # modify this here!. 
            boxes_ii = vid_bboxes[ii].copy()
            # set prob dependent on the size of the boxes. (if 5D then take the first as a measure of objectness else assume 1.)
            if boxes_ii.shape[-1] == 4: 
                prob_ii = np.ones(len(boxes_ii)) 
            else:
                prob_ii = boxes_ii[:,0].copy()
                boxes_ii = boxes_ii[:,1:].copy()

            boxes_ii = boxes_ii / float(ds_factor)
            
            # clip
            boxes_ii[:,0] = np.clip(boxes_ii[:,0], 0, im_shape[1]-1)
            boxes_ii[:,1] = np.clip(boxes_ii[:,1], 0, im_shape[0]-1)
            boxes_ii[:,2] = np.clip(boxes_ii[:,2], 0, im_shape[1]-1)
            boxes_ii[:,3] = np.clip(boxes_ii[:,3], 0, im_shape[0]-1)
        
            # remove all boxes that are have an area 1 pixels or less.               
            boxes_ii_w = boxes_ii[:,2] - boxes_ii[:,0]
            boxes_ii_h = boxes_ii[:,3] - boxes_ii[:,1]
            boxes_ii = boxes_ii[boxes_ii_w*boxes_ii_h>0]

            # suppress implausible. 
            if remove_eccentric_bbox:
              boxes_ii, keep_ind_ii = remove_very_large_bbox(boxes_ii, 
                                                            im_shape, 
                                                            thresh=0.5, 
                                                            aspect_ratio = 3, # we don't expect it to be very long. 
                                                            method='density', 
                                                            max_density=max_dense_bbox_cover, 
                                                            mode='fast', 
                                                            return_ind=True)
            else:
              keep_ind_ii = np.ones(len(boxes_ii))>0

            assert(len(boxes_ii) == len(keep_ind_ii)) # keep a check here.       
            prob_ii = prob_ii[keep_ind_ii]

            for jj in np.arange(len(boxes_ii)):
                vid_bbox_tracks.append([[ii, prob_ii[jj], boxes_ii[jj]]]) # add the prob in as another entry here. 
                vid_bbox_tracks_last_time.append(ii) # update with current time. 
                vid_match_check_bbox_tracks.append([[ii, 1]]) # update with the current time., this is just to select whether this box was propagated or found in the original detections. 
                
        """
        load the current working set of tracks. 
        """
        # 1) check for track termination 
        # get the last timepoint. 
        boxes_ii_track_time = np.hstack(vid_bbox_tracks_last_time)
        
        # check if any of the box tracks need to be terminated. 
        track_terminate =  boxes_ii_track_time < ii - wait_time
        track_terminate_id = np.arange(len(track_terminate))[track_terminate]

        if len(track_terminate_id)>0:
            # update the relevant info.
            for idd in track_terminate_id:
                #update
                terminated_vid_bbox_tracks.append(vid_bbox_tracks[idd][:-wait_time-1])
                terminated_check_match_bbox_tracks.append(vid_match_check_bbox_tracks[idd][:-wait_time-1])
                # terminated_vid_cell_tracks_properties.append(vid_cell_tracks_properties[idd][:-wait_time-1])

            # # reform the working set. 
            # vid_cell_tracks = [vid_cell_tracks[jjj] for jjj in np.arange(len(vid_cell_tracks)) if jjj not in track_terminate_id] # i guess this is to keep track of actual cell ids that we segmented.
            vid_bbox_tracks = [vid_bbox_tracks[jjj] for jjj in np.arange(len(vid_bbox_tracks)) if jjj not in track_terminate_id] # this is to keep track of the actual bbox we use, including inferred. 
            # vid_cell_tracks_properties = [vid_cell_tracks_properties[jjj] for jjj in np.arange(len(vid_cell_tracks_properties)) if jjj not in track_terminate_id]
            vid_bbox_tracks_last_time = [vid_bbox_tracks_last_time[jjj] for jjj in np.arange(len(vid_bbox_tracks_last_time)) if jjj not in track_terminate_id]
            vid_match_check_bbox_tracks = [vid_match_check_bbox_tracks[jjj] for jjj in np.arange(len(vid_match_check_bbox_tracks)) if jjj not in track_terminate_id]

        # load the current version of the boxes. 
        boxes_ii_track = np.array([bb[-1][-1] for bb in vid_bbox_tracks]) # the bboxes to consider.
        boxes_ii_track_prob = np.array([bb[-1][1] for bb in vid_bbox_tracks]) 
        boxes_ii_track_time = np.array([bb[-1][0] for bb in vid_bbox_tracks]) # the time of the last track. 

        """
        Infer the next frame boxes from current boxes using optical flow.    
        """
        boxes_ii_track_pred = []
        
        for jjj in np.arange(len(boxes_ii_track)):
            """
            Predict using the flow the likely position of boxes. 
            """
            new_tfs, boxes_ii_pred = predict_new_boxes_flow_tf(boxes_ii_track[jjj][None,:], 
                                                               vid_flow[boxes_ii_track_time[jjj]])
            # clip  
            boxes_ii_pred[:,0] = np.clip(boxes_ii_pred[:,0], 0, im_shape[1]-1)
            boxes_ii_pred[:,1] = np.clip(boxes_ii_pred[:,1], 0, im_shape[0]-1)
            boxes_ii_pred[:,2] = np.clip(boxes_ii_pred[:,2], 0, im_shape[1]-1)
            boxes_ii_pred[:,3] = np.clip(boxes_ii_pred[:,3], 0, im_shape[0]-1)
            boxes_ii_track_pred.append(boxes_ii_pred[0])
  
        boxes_ii_track_pred = np.array(boxes_ii_track_pred)
            
        """
        load the next frame boxes. which are the candidates.  
        """
        # bboxfile_jj = vid_bboxes[ii+1]
        # prob_jj, boxes_jj = load_bbox_frame_voc( vid[ii+1], 
                                                 # bboxfile_jj)
        boxes_jj = vid_bboxes[ii+1].copy();
        # set prob dependent on the size of the boxes. (if 5D then take the first as a measure of objectness else assume 1.)
        if boxes_jj.shape[-1] == 4: 
            prob_jj = np.ones(len(boxes_jj))
        else:
            prob_jj = boxes_jj[:,0].copy()
            boxes_jj = boxes_jj[:,1:].copy()
        
        boxes_jj = boxes_jj/float(ds_factor)

        # clip
        boxes_jj[:,0] = np.clip(boxes_jj[:,0], 0, im_shape[1]-1)
        boxes_jj[:,1] = np.clip(boxes_jj[:,1], 0, im_shape[0]-1)
        boxes_jj[:,2] = np.clip(boxes_jj[:,2], 0, im_shape[1]-1)
        boxes_jj[:,3] = np.clip(boxes_jj[:,3], 0, im_shape[0]-1) 
        # remove all boxes that are have an area 1 pixels or less.               
        boxes_jj_w = boxes_jj[:,2] - boxes_jj[:,0]
        boxes_jj_h = boxes_jj[:,3] - boxes_jj[:,1]
        boxes_jj = boxes_jj[boxes_jj_w*boxes_jj_h>0]

        # suppress implausible. 
        if remove_eccentric_bbox:
          boxes_jj, keep_ind_jj = remove_very_large_bbox(boxes_jj, 
                                                      im_shape, 
                                                      thresh=0.5, 
                                                      aspect_ratio = 3, # we don't expect it to be very long. 
                                                      method='density', 
                                                      max_density=max_dense_bbox_cover, 
                                                      mode='fast', return_ind=True)
        else:
          keep_ind_jj = np.ones(len(boxes_jj))>0
              
        prob_jj = prob_jj[keep_ind_jj]
        
        """
        build the association matrix and match boxes based on iou. 
        """
        iou_matrix = bbox_iou_corner_xy(boxes_ii_track_pred, 
                                        boxes_jj); 
        iou_matrix = np.clip(1.-iou_matrix, 0, 1) # to make it a dissimilarity matrix. 
        
        # solve the pairing problem.
        ind_ii, ind_jj = linear_sum_assignment(iou_matrix)
        
        # threshold as the matching is maximal. 
        iou_ii_jj = iou_matrix[ind_ii, ind_jj].copy()
        keep = iou_ii_jj <= (1-iou_match)
        # keep = iou_ii_jj <= dist_thresh
        ind_ii = ind_ii[keep>0]; 
        ind_jj = ind_jj[keep>0]
        
        """
        Update the trajectories. 
        """
        # update first the matched.
        for match_ii in np.arange(len(ind_ii)):
            # vid_cell_tracks[ind_ii[match_ii]].append([ii+1, ind_jj[match_ii]]) # i guess this is to keep track of actual cell ids that we segmented.
            vid_bbox_tracks[ind_ii[match_ii]].append([ii+1, prob_jj[ind_jj[match_ii]], boxes_jj[ind_jj[match_ii]]]) # this is to keep track of the actual bbox we use, including inferred. 
            # vid_cell_tracks_properties[ind_ii[match_ii]].append([ii+1, masks_jj_properties[ind_jj[match_ii]]]) # append the properties of the next time point. 
            vid_bbox_tracks_last_time[ind_ii[match_ii]] = ii+1 # let this be a record of the last time a 'real' segmentation was matched, not one inferred from optical flow. 
            # vid_mask_tracks_last[ind_ii[match_ii]] = masks_jj[...,ind_jj[match_ii]] # retain just the last masks thats relevant for us. 
            # vid_mask_tracks_last[ind_ii[match_ii]] = boxes_jj[ind_jj[match_ii]]
            vid_match_check_bbox_tracks[ind_ii[match_ii]].append([ii+1, 1]) # success, append 0 
            
        no_match_ind_ii = np.setdiff1d(np.arange(len(boxes_ii_track_pred)), ind_ii)
        no_match_ind_jj = np.setdiff1d(np.arange(len(boxes_jj)), ind_jj)
        
        for idd in no_match_ind_ii:
            # these tracks already exist so we just add to the existant tracks.         
            # vid_cell_tracks[idd].append([ii+1, -1]) # i guess this is to keep track of actual cell ids that we segmented.
            vid_bbox_tracks[idd].append([ii+1, boxes_ii_track_prob[idd], boxes_ii_track_pred[idd]]) # this is to keep track of the actual bbox we use, including inferred. 
            # vid_cell_tracks_properties[idd].append([ii+1, properties_ii_track_pred[idd]])
            # vid_bbox_tracks_last_time[idd] = ii+1 # let this be a record of the last time a 'real' segmentation was matched, not one inferred from optical flow. 
            # vid_mask_tracks_last[idd] = boxes_ii_track_pred[idd] # retain just the last masks thats relevant for us.
            vid_match_check_bbox_tracks[idd].append([ii+1, 0]) # no success, append 0  
    
        for idd in no_match_ind_jj:
            # these tracks don't exsit yet so we need to create new tracks. 
            # vid_cell_tracks.append([[ii+1, idd]]) # i guess this is to keep track of actual cell ids that we segmented.
            vid_bbox_tracks.append([[ii+1, prob_jj[idd], boxes_jj[idd]]]) # this is to keep track of the actual bbox we use, including inferred. 
            # vid_cell_tracks_properties.append([[ii+1, masks_jj_properties[idd]]])
            vid_bbox_tracks_last_time.append(ii+1) # let this be a record of the last time a 'real' segmentation was matched, not one inferred from optical flow. 
            # vid_mask_tracks_last.append(masks_jj[...,idd]) # retain just the last masks thats relevant for us. 
            # vid_mask_tracks_last.append(boxes_jj[idd])
            vid_match_check_bbox_tracks.append([[ii+1, 1]])
                
# =============================================================================
#     Combine the tracks togther
# =============================================================================
    vid_bbox_tracks_all = terminated_vid_bbox_tracks + vid_bbox_tracks # combine
    vid_match_checks_all = terminated_check_match_bbox_tracks + vid_match_check_bbox_tracks


# =============================================================================
#     Compute some basic track properties for later filtering. 
# =============================================================================    
    vid_bbox_tracks_all_lens = np.hstack([len(tra) for tra in vid_bbox_tracks_all])
    vid_bbox_tracks_all_start_time = np.hstack([tra[0][0] for tra in vid_bbox_tracks_all])
    
    print(vid_bbox_tracks_all_lens)
    print(vid_bbox_tracks_all_start_time)
    
    vid_bbox_tracks_lifetime_ratios = vid_bbox_tracks_all_lens / (len(vid_flow) - vid_bbox_tracks_all_start_time).astype(np.float32)
    
# =============================================================================
#     Turn into a proper array of n_tracks x n_time x 4 or 5.... 
# =============================================================================
    vid_bbox_tracks_prob_all_array, vid_bbox_tracks_all_array = bbox_tracks_2_array(vid_bbox_tracks_all, nframes=nframes, ndim=boxes_ii.shape[1])
    vid_match_checks_all_array = bbox_scalar_2_array(vid_match_checks_all, nframes=nframes)

# =============================================================================
#   Apply the given filtering parameters for visualization else it will be messy.        
# =============================================================================

    if to_viz:

        fig, ax = plt.subplots()
        ax.scatter(vid_bbox_tracks_all_start_time, 
                   vid_bbox_tracks_all_lens, 
                   c=vid_bbox_tracks_lifetime_ratios, 
                   vmin=0, 
                   vmax=1, cmap='coolwarm')
        plt.show()


        # keep filter
        keep_filter = np.logical_and(vid_bbox_tracks_all_lens>=min_tra_len_filter, 
                                     vid_bbox_tracks_lifetime_ratios>=min_tra_lifetime_ratio)
        keep_ids = np.arange(len(vid_bbox_tracks_all_lens))[keep_filter]

        plot_vid_bbox_tracks_all = vid_bbox_tracks_all_array[keep_ids].copy()
        plot_colors = sns.color_palette('hls', len(plot_vid_bbox_tracks_all))

        # print(len(plot_vid_bbox_tracks_all), len(vid_bbox_tracks_all_array))
        # iterate over time. 
        for frame_no in np.arange(nframes):

            fig, ax = plt.subplots(figsize=(5,5))
            plt.title('Frame: %d' %(frame_no+1))
            
            if vid is not None:
                vid_overlay = vid[frame_no].copy()
            else:
                vid_overlay = np.zeros(im_shape)

            ax.imshow(vid_overlay, cmap='gray')
            
            for ii in np.arange(len(plot_vid_bbox_tracks_all))[:]:
                bbox_tra_ii = plot_vid_bbox_tracks_all[ii][frame_no] # fetch it at this point in time. 

                if ~np.isnan(bbox_tra_ii[0]):
                    # then there is a valid bounding box. 
                    x1,y1,x2,y2 = bbox_tra_ii 
                    ax.plot( [x1,x2,x2,x1,x1], 
                             [y1,y1,y2,y2,y1], lw=1, color = plot_colors[ii])

            ax.set_xlim([0, im_shape[1]-1])
            ax.set_ylim([im_shape[0]-1, 0])
            plt.axis('off')
            plt.grid('off')
            
            if saveanalysisfolder is not None:
                fig.savefig(os.path.join(saveanalysisfolder_movie_bbox, 'Frame-%s.png' %(str(frame_no).zfill(3))), bbox_inches='tight')
            plt.show()
            plt.close(fig)
        
    # spio.savemat(savematfile, 
    #                 {'boundaries':organoid_boundaries, 
    #                 'initial_seg': seg0_grab,
    #                 'motion_source_map': spixels_B_motion_sources,
    #                 'seg_pts': organoid_segs_pts})

    # return organoid_boundaries, organoid_segs_pts
    return vid_bbox_tracks_prob_all_array, vid_bbox_tracks_all, vid_bbox_tracks_all_array, vid_match_checks_all_array, (vid_bbox_tracks_all_lens, vid_bbox_tracks_all_start_time, vid_bbox_tracks_lifetime_ratios)


def tracks2Darray_to_3Darray(tracks2D_time, uv_params_time):

    from ..Unzipping import unzip_new as uzip
    import numpy as np 

    m, n = uv_params_time.shape[1:-1]
    n_tracks, n_time, _ = tracks2D_time.shape
    tracks3D_time = np.zeros((n_tracks, n_time, 3))
    tracks3D_time[:] = np.nan 

    for tt in np.arange(n_time):
        # iterate over frame
        uv_params = uv_params_time[tt].copy()
        tracks2D_tt = tracks2D_time[:,tt].copy()

        # only interpolate for non nan vals. 
        val_track = np.logical_not(np.isnan(tracks2D_tt[:,0]))

        pts2D = tracks2D_tt[val_track>0].copy()
        pts2D[...,0] = np.clip(pts2D[...,0], 0, n-1) # this assumes x,y... 
        pts2D[...,1] = np.clip(pts2D[...,1], 0, m-1)

        # is this interpolation bad? 
        pts3D = np.array([uzip.map_intensity_interp2(pts2D[:,::-1], 
                                                     grid_shape=(m,n), 
                                                     I_ref=uv_params[...,ch], 
                                                     method='linear', 
                                                     cast_uint8=False, s=0) for ch in np.arange(uv_params.shape[-1])]).T
        tracks3D_time[val_track>0, tt, :] = pts3D.copy()

    return tracks3D_time



# =============================================================================
#   Add some track postprocessing support namely removal of all nan tracks and non-max suppression amongst tracks. (operating purely on bboxes!)
# =============================================================================

def calculate_iou_matrix_time(box_arr1, box_arr2, eps=1e-9):
        
    import numpy as np 
    x11 = box_arr1[...,0]; y11 = box_arr1[...,1]; x12 = box_arr1[...,2]; y12 = box_arr1[...,3]
    x21 = box_arr2[...,0]; y21 = box_arr2[...,1]; x22 = box_arr2[...,2]; y22 = box_arr2[...,3]
    m,n = x11.shape
    # # n_tracks x n_time. 
    # flip this.
    x11 = x11.T; y11 = y11.T; x12 = x12.T; y12 = y12.T
    x21 = x21.T; y21 = y21.T; x22 = x22.T; y22 = y22.T 

    xA = np.maximum(x11[...,None], x21[:,None,:])
    yA = np.maximum(y11[...,None], y21[:,None,:])
    xB = np.minimum(x12[...,None], x22[:,None,:])
    yB = np.minimum(y12[...,None], y22[:,None,:])
    
    interArea = np.maximum((xB - xA + eps), 0) * np.maximum((yB - yA + eps), 0)
    boxAArea = (x12 - x11 + eps) * (y12 - y11 + eps)
    boxBArea = (x22 - x21 + eps) * (y22 - y21 + eps)

    # iou = interArea / (boxAArea + np.transpose(boxBArea) - interArea)
    iou = interArea / (boxAArea[...,None] + boxBArea[:,None,:] - interArea)
    # # iou = iou.reshape((m,n))
    return iou

# replace with iou bbox_tracks
def iou_boundary_tracks(tra_1, tra_2):

    import numpy as np 

    n_pts = len(tra_1) 
    iou_time = np.zeros(n_pts)

    for ii in range(n_pts):
        tra_1_ii = tra_1[ii]
        tra_2_ii = tra_2[ii]

        if np.isnan(tra_1_ii[0,0]) or np.isnan(tra_2_ii[0,0]):
            iou_time[ii] = np.nan
        else:
            x1, x2 = np.min(tra_1_ii[...,1]), np.max(tra_1_ii[...,1])
            y1, y2 = np.min(tra_1_ii[...,0]), np.max(tra_1_ii[...,0])
            
            x1_, x2_ = np.min(tra_2_ii[...,1]), np.max(tra_2_ii[...,1])
            y1_, y2_ = np.min(tra_2_ii[...,0]), np.max(tra_2_ii[...,0])
            
            bbox1 = np.hstack([x1,y1,x2,y2])
            bbox2 = np.hstack([x1_,y1_,x2_,y2_])
            
            print(bbox1, bbox2)
            iou_12 = get_iou(bbox1, bbox2)
            iou_time[ii] = iou_12

    return iou_time

def get_iou(bb1, bb2):
    """
    Calculate the Intersection over Union (IoU) of two bounding boxes.

    Parameters
    ----------
    bb1 : dict
        Keys: {'x1', 'x2', 'y1', 'y2'}
        The (x1, y1) position is at the top left corner,
        the (x2, y2) position is at the bottom right corner
    bb2 : dict
        Keys: {'x1', 'x2', 'y1', 'y2'}
        The (x, y) position is at the top left corner,
        the (x2, y2) position is at the bottom right corner

    Returns
    -------
    float
        in [0, 1]
    """
    if np.sum(bb1) < 1 and np.sum(bb2) < 1:
        iou = 0
    else:
        bb1 = {'x1': bb1[0], 
               'y1': bb1[1], 
               'x2': bb1[2],
               'y2': bb1[3]}
        bb2 = {'x1': bb2[0], 
               'y1': bb2[1], 
               'x2': bb2[2],
               'y2': bb2[3]}
        
    #    print(bb1)
    #    print(bb2)
    #     test for 0
            
        # allow for collapsed boxes.           
        assert bb1['x1'] <= bb1['x2']
        assert bb1['y1'] <= bb1['y2']
        assert bb2['x1'] <= bb2['x2']
        assert bb2['y1'] <= bb2['y2']
    
        # determine the coordinates of the intersection rectangle
        x_left = max(bb1['x1'], bb2['x1'])
        y_top = max(bb1['y1'], bb2['y1'])
        x_right = min(bb1['x2'], bb2['x2'])
        y_bottom = min(bb1['y2'], bb2['y2'])
    
        if x_right < x_left or y_bottom < y_top:
            return 0.0
    
        # The intersection of two axis-aligned bounding boxes is always an
        # axis-aligned bounding box
        intersection_area = (x_right - x_left) * (y_bottom - y_top)
    
        # compute the area of both AABBs
        bb1_area = (bb1['x2'] - bb1['x1']) * (bb1['y2'] - bb1['y1'])
        bb2_area = (bb2['x2'] - bb2['x1']) * (bb2['y2'] - bb2['y1'])
    
        # compute the intersection over union by taking the intersection
        # area and dividing it by the sum of prediction + ground-truth
        # areas - the interesection area
        iou = intersection_area / float(bb1_area + bb2_area - intersection_area)
        assert iou >= 0.0
        assert iou <= 1.0
    return iou

def iou_bbox_tracks(tra_1, tra_2):

    import numpy as np 

    n_pts = len(tra_1) 
    iou_time = np.zeros(n_pts)

    for ii in range(n_pts):
        tra_1_ii = tra_1[ii] # get the track at a particular timepoint should be 4
        tra_2_ii = tra_2[ii]

        if np.isnan(tra_1_ii[0,0]) or np.isnan(tra_2_ii[0,0]):
            iou_time[ii] = np.nan # prop the nan onwards. 
        else:
            x1, y1, x2, y2 = tra_1_ii
            x1_, y1_, x2_, y2_ = tra_2_ii
            # x1, x2 = np.min(tra_1_ii[...,1]), np.max(tra_1_ii[...,1])
            # y1, y2 = np.min(tra_1_ii[...,0]), np.max(tra_1_ii[...,0])
            
            # x1_, x2_ = np.min(tra_2_ii[...,1]), np.max(tra_2_ii[...,1])
            # y1_, y2_ = np.min(tra_2_ii[...,0]), np.max(tra_2_ii[...,0])
            bbox1 = np.hstack([x1,y1,x2,y2])
            bbox2 = np.hstack([x1_,y1_,x2_,y2_])
            
            # print(bbox1, bbox2)
            iou_12 = get_iou(bbox1, bbox2)
            iou_time[ii] = iou_12

    return iou_time


def pairwise_iou_tracks(boundaries_list):
    
    import itertools
    import numpy as np 
    
    Ns = np.hstack([len(bb) for bb in boundaries_list]) # this is for dissecting IDs.
    
    ind_ids = np.hstack([[str(jj)+'_'+str(ii) for jj in range(Ns[ii])] for ii in range(len(Ns))])
    # stack all of them together. 
    all_boundaries = np.vstack(boundaries_list)
    
    # print(all_boundaries.shape)
    sim_matrix = np.zeros((len(all_boundaries), len(all_boundaries)))
    shared_time_matrix = np.zeros((len(all_boundaries), len(all_boundaries)))
    
    for i, j in itertools.combinations(range(len(all_boundaries)),2):
        
        boundary_i = all_boundaries[i]
        boundary_j = all_boundaries[j]
        
        # iou_track = iou_boundary_tracks(boundary_i, boundary_j)
        iou_track = iou_bbox_tracks(boundary_i, boundary_j)
        
        sim_matrix[i,j] = np.nanmean(iou_track)
        sim_matrix[j,i] = np.nanmean(iou_track)
        
        shared_time_matrix[i,j] = ~np.isnan(iou_track).sum()
        shared_time_matrix[j,i] = ~np.isnan(iou_track).sum()
        
    return ind_ids, (sim_matrix, shared_time_matrix)


def pairwise_iou_tracks_fast(boundaries_list, eps=1e-9, return_bbox=False, avg_func=np.nanmean):
        
    import numpy as np 
    import time 
    
    Ns = np.hstack([len(bb) for bb in boundaries_list]) # this is for dissecting IDs.
    
    ind_ids = np.hstack([[str(jj)+'_'+str(ii) for jj in range(Ns[ii])] for ii in range(len(Ns))])
    # stack all of them together. 
    all_boundaries = np.vstack(boundaries_list) # flatten all. 
    
    # # turn the boundaries into bbox. 
    # all_boundaries_bbox_xmin = np.min(all_boundaries[...,0], axis=-1)
    # all_boundaries_bbox_xmax = np.max(all_boundaries[...,2], axis=-1)
    # all_boundaries_bbox_ymin = np.min(all_boundaries[...,1], axis=-1)
    # all_boundaries_bbox_ymax = np.max(all_boundaries[...,3], axis=-1)
    
    # all_boundaries_bbox = np.concatenate([all_boundaries_bbox_xmin[...,None], 
    #                                       all_boundaries_bbox_ymin[...,None], 
    #                                       all_boundaries_bbox_xmax[...,None], 
    #                                       all_boundaries_bbox_ymax[...,None]], axis=-1)
    all_boundaries_bbox = all_boundaries.copy()
    
    all_sim_matrix = calculate_iou_matrix_time(all_boundaries_bbox, 
                                                all_boundaries_bbox, 
                                            eps=eps)
    
    sim_matrix = avg_func(all_sim_matrix, axis=0)
    shared_time_matrix = np.sum(~np.isnan(all_sim_matrix), axis=0)
     
    if return_bbox:
        return ind_ids, (sim_matrix, shared_time_matrix), (all_boundaries_bbox, sim_matrix)
    else:
        return ind_ids, (sim_matrix, shared_time_matrix)

def unique_pts(a):
    import numpy as np 
    # returns the unique rows of an array. 
    return np.vstack(list({tuple(row) for row in a}))

def get_id_cliques(id_list):
    """
    given a list of identities, merges the common ids into a cluster/clique
    """
    N = len(id_list)
    cliques = [id_list[0]]
    
    for ii in range(1, N):
        
        id_list_ii = id_list[ii]
        
        add_clique = True
        for cc_i, cc in enumerate(cliques):
            if len(np.intersect1d(id_list_ii, cc)) > 0:
                cliques[cc_i] = np.unique(np.hstack([cc, id_list_ii]))
                add_clique = False
                break
            
        if add_clique:
            cliques.append(id_list_ii)
            
    return cliques

# of the two versions, this is more useful because it ensures more temporal continuity.
def objectness_score_tracks_time(track, vid_score, mean_func=np.nanmean): 
    
    # give an image and a timepoint which the image corresponds to, to get the objectness score. default uses the first timepoint of the video.  
    from skimage.draw import polygon 
    
    img_shape = vid_score[0].shape
    nframes = len(track)
    scores = []
    
    for frame in range(nframes):
        coords = track[frame].copy()
        print(coords)
        if ~np.isnan(coords[0]):
            x1,y1,x2,y2 = coords
            coords = np.vstack([[y1,y1,y2,y2],
                                [x1,x2,x2,x1]])
            rr,cc = polygon(coords[:,0], 
                            coords[:,1], shape=img_shape)
            score = mean_func(vid_score[frame][rr,cc])
            scores.append(score)
        else:
            scores.append(0)
    
    score = np.nanmean(scores)
    
    return score


def track_iou_time(track, use_bbox=True): 
    
    """
    frame to frame iou score 
    """
    nframes = len(track)
    first_diff_iou = []
    
    for frame_no in np.arange(nframes-1):
        
        tra_1 = track[frame_no]
        tra_2 = track[frame_no+1]
        
        if np.isnan(tra_1[0]) or np.isnan(tra_2[0]):
            first_diff_iou.append(np.nan)
        else:
            
            if use_bbox:
                x1, y1, x2, y2 = tra_1
                x1_, y1_, x2_, y2_ = tra_2
                # x1, x2 = np.min(tra_1[...,1]), np.max(tra_1[...,1])
                # y1, y2 = np.min(tra_1[...,0]), np.max(tra_1[...,0])
                
                # x1_, x2_ = np.min(tra_2[...,1]), np.max(tra_2[...,1])
                # y1_, y2_ = np.min(tra_2[...,0]), np.max(tra_2[...,0])
                bbox1 = np.hstack([x1,y1,x2,y2])
                bbox2 = np.hstack([x1_,y1_,x2_,y2_])
                
                # print(bbox1, bbox2)
                iou_12 = get_iou(bbox1, bbox2)
                first_diff_iou.append(iou_12)
    
    first_diff_iou = np.hstack(first_diff_iou)
    
    return first_diff_iou


# the following bypasses the need for point alignment along the tracks? by using area. 
def smoothness_score_tracks_iou(track, mean_func=np.nanmean, second_order=False, use_bbox=True): 
    
    # get the second order track differences
    # from skimage.draw import polygon 
    import numpy as np 
        
    first_diff_iou = track_iou_time(track, use_bbox=use_bbox)
        
    if second_order:
        second_diff_norm = np.gradient(first_diff_iou)
    else:
        second_diff_norm = first_diff_iou.copy()
    # second_diff_norm = np.nanmean(second_diff_norm, axis=1) # obtain the mean over the boundaries., should we use median?

    if second_order:
        score = -mean_func(second_diff_norm) # this should be maxed for iou.
    else:
        score = mean_func(second_diff_norm)
    return score

def nan_stability_score_tracks(track): 
    """
    simply the fraction of frames for which we were able to successfully track the object. 
    """     
    len_track = float(len(track))
    valid_track_times = np.sum(~np.isnan(track[:,0]))

    score = valid_track_times / len_track 
    
    return score


def non_stable_track_suppression_filter(obj_vid, 
                                        org_tracks_list, # use the bbox directly.  
                                        track_overlap_thresh=0.25, 
                                        weight_nan=1., weight_smooth=0.1, max_obj_frames=5,
                                        obj_mean_func=np.nanmean,
                                        smoothness_mean_func=np.nanmean,
                                        fast_comp=True,
                                        debug_viz=False):
    """
    Combines the utility codes above to filter the raw organoid boundary tracks. extends to single and multiple channels.  
        need to absolutely change this... 

        obj_vid: the metric to score objectness .... 
    """
    if debug_viz:
        import pylab as plt 
    
    # 1. apply iou calculations pairwise between tracks to score overlap across channels
    if fast_comp:
        ind_ids, (sim_matrix, shared_time_matrix) = pairwise_iou_tracks_fast(org_tracks_list)
        # detrend the diagonals.(which is self connections)
        sim_matrix = sim_matrix - np.diag(np.diag(sim_matrix)) 
    else:
        ind_ids, (sim_matrix, shared_time_matrix) = pairwise_iou_tracks(org_tracks_list) # this concatenates all the tracks etc together, resolving all inter-, intra- overlaps

    sim_matrix_ = sim_matrix.copy()
    sim_matrix_[np.isnan(sim_matrix)] = 0 # replace any nan values which will not be useful. 

    # 2. detect overlaps and cliques (clusters of tracks that correspond to one dominant organoid)
    tracks_overlap = np.where(sim_matrix_ >= track_overlap_thresh)

#    print(tracks_overlap)
    if len(tracks_overlap[0]) > 0:

        # 2b. if there is any evidence of overlap! we collect this all together.  
        overlap_positive_inds = np.vstack(tracks_overlap).T
        overlap_positive_inds = np.sort(overlap_positive_inds, axis=1)
#        print(overlap_positive_inds)
        # overlap_positive_inds = np.unique(overlap_positive_inds, axis=0) #remove duplicate rows.
        overlap_positive_inds = unique_pts(overlap_positive_inds)
        
        # merge these indices to identify unique cliques ... -> as those will likely be trying to track the same organoid. 
        cliq_ids = get_id_cliques(overlap_positive_inds) 
  
        # 3. clique resolution -> determining which organoid is actually being tracked, and which track offers best tracking performance on average from all candidates in the clique. 
        assigned_cliq_track_ids = [] # stores which of the tracks we should use from the overlapped channels.
    
        for cc in cliq_ids[:]:

            # iterate, use objectness score provided by the input vid, to figure out which organoid is being tracked. 
            ind_ids_cc = ind_ids[cc] # what are the possible ids here. 

            obj_score_cc = []
            tra_stable_scores_cc = []
            
            # in the order of organoid id and channel. 
            if debug_viz:
                import seaborn as sns

                ch_colors = sns.color_palette('Set1', len(org_tracks_list))

                fig, ax = plt.subplots()
                ax.imshow(obj_vid[0], alpha=0.5) # just visualise the first frame is enough.
            
            for ind_ids_ccc in ind_ids_cc:
                org_id, org_ch = ind_ids_ccc.split('_')
                org_id = int(org_id)
                org_ch = int(org_ch)
                
                boundary_org = org_tracks_list[org_ch][org_id]

                # this is the problem? 
                # objectness score for deciding the dominant organoid. 
                obj_score = objectness_score_tracks_time(boundary_org[:max_obj_frames], 
                                                          obj_vid[:max_obj_frames,...,org_ch], 
                                                              mean_func=obj_mean_func)
                obj_score_cc.append(obj_score)
                
                # stability score which is weighted on 2 factors. to determine which track. 
                nan_score = nan_stability_score_tracks(boundary_org)
                
                # this should be a minimisation ..... ! 
                # smooth_score = smoothness_score_tracks(boundary_org, 
                #                                       mean_func=np.nanmean)
                smooth_score = smoothness_score_tracks_iou(boundary_org, 
                                                           mean_func=smoothness_mean_func)
                
                tra_stable_scores_cc.append(weight_nan*nan_score+weight_smooth*smooth_score)

            if debug_viz:
                ax.set_title( 'org: %s, stable: %s'  %(ind_ids_cc[np.argmax(obj_score_cc)], 
                                                ind_ids_cc[np.argmax(tra_stable_scores_cc)]))
                plt.show()
            
            # stack all the scores.     
            obj_score_cc = np.hstack(obj_score_cc)
            tra_stable_scores_cc = np.hstack(tra_stable_scores_cc)

            # decide on the organoid and track (argmax)
            cliq_org_id_keep = ind_ids_cc[np.argmax(obj_score_cc)]
            cliq_track_id_keep = ind_ids_cc[np.argmax(tra_stable_scores_cc)]
            
            # save this out for processing. 
            assigned_cliq_track_ids.append([cliq_org_id_keep, cliq_track_id_keep])

        # 4. new org_tracks_list production based on the filtered information. 
        org_tracks_list_out = []

        for list_ii in range(len(org_tracks_list)):

            org_tracks_list_ii = org_tracks_list[list_ii]
            org_tracks_list_ii_out = []

            for org_ii in range(len(org_tracks_list_ii)):
                tra_int_id = str(org_ii)+'_'+str(list_ii) # create the string id lookup. 
                
                include_track = True

                for cliq_ii in range(len(cliq_ids)):
                    ind_ids_cc = ind_ids[cliq_ids[cliq_ii]] # gets the clique members in string form -> is this organoid part of a clique. 
                    
                    if tra_int_id in ind_ids_cc:
                        include_track = False # do not automatically include.  

                        # test is this the dominant organoid in the clique. 
                        cliq_organoid_assign, cliq_organoid_assign_track = assigned_cliq_track_ids[cliq_ii] # get the assignment information of the clique. 
                        
                        if tra_int_id == cliq_organoid_assign:
                            # if this is the dominant organoid then we add the designated track. 
                            org_id_tra_assign, org_ch_tra_assign = cliq_organoid_assign_track.split('_')
                            org_id_tra_assign = int(org_id_tra_assign); org_ch_tra_assign=int(org_ch_tra_assign)
                            
                            org_tracks_list_ii_out.append(org_tracks_list[org_ch_tra_assign][org_id_tra_assign])

                        # do nothing otherwise -> exclude this organoid basically. 
                
                if include_track: 
                    # directly include. 
                    org_tracks_list_ii_out.append(org_tracks_list_ii[org_ii])

            if len(org_tracks_list_ii_out) > 0:
                # stack the tracks. 
                org_tracks_list_ii_out = np.array(org_tracks_list_ii_out)
            
            org_tracks_list_out.append(org_tracks_list_ii_out)
    else:
        org_tracks_list_out = list(org_tracks_list)

    # cleaned tracks, in the same input format as a list of numpy tracks for each channel. 
    return org_tracks_list_out

def non_maxlen_track_suppression_filter(obj_vid, 
                                        org_tracks_list, # use the bbox directly.  
                                        track_overlap_thresh=0.25, 
                                        max_obj_frames=5,
                                        obj_mean_func=np.nanmean,
                                        fast_comp=True,
                                        debug_viz=False):
    """
    Combines the utility codes above to filter the raw organoid boundary tracks. extends to single and multiple channels.  
        need to absolutely change this... 

        obj_vid: the metric to score objectness .... 
    """
    if debug_viz:
        import pylab as plt 
    
    # 1. apply iou calculations pairwise between tracks to score overlap across channels - we only suppress. those that share a lot of overlap. 
    if fast_comp:
        ind_ids, (sim_matrix, shared_time_matrix) = pairwise_iou_tracks_fast(org_tracks_list)
        # detrend the diagonals.(which is self connections)
        sim_matrix = sim_matrix - np.diag(np.diag(sim_matrix)) 
    else:
        ind_ids, (sim_matrix, shared_time_matrix) = pairwise_iou_tracks(org_tracks_list) # this concatenates all the tracks etc together, resolving all inter-, intra- overlaps

    sim_matrix_ = sim_matrix.copy()
    sim_matrix_[np.isnan(sim_matrix)] = 0 # replace any nan values which will not be useful. 

    # 2. detect overlaps and cliques (clusters of tracks that correspond to one dominant organoid)
    tracks_overlap = np.where(sim_matrix_ >= track_overlap_thresh)

#    print(tracks_overlap)
    if len(tracks_overlap[0]) > 0:

        # 2b. if there is any evidence of overlap! we collect this all together.  
        overlap_positive_inds = np.vstack(tracks_overlap).T
        overlap_positive_inds = np.sort(overlap_positive_inds, axis=1)
#        print(overlap_positive_inds)
        # overlap_positive_inds = np.unique(overlap_positive_inds, axis=0) #remove duplicate rows.
        overlap_positive_inds = unique_pts(overlap_positive_inds)
        
        # merge these indices to identify unique cliques ... -> as those will likely be trying to track the same organoid. 
        cliq_ids = get_id_cliques(overlap_positive_inds) 
  
        # 3. clique resolution -> determining which organoid is actually being tracked, and which track offers best tracking performance on average from all candidates in the clique. 
        assigned_cliq_track_ids = [] # stores which of the tracks we should use from the overlapped channels.
    
        for cc in cliq_ids[:]:

            # iterate, use objectness score provided by the input vid, to figure out which organoid is being tracked. 
            ind_ids_cc = ind_ids[cc] # what are the possible ids here. 
            print(ind_ids_cc)

            obj_score_cc = []
            tra_stable_scores_cc = []
            
            # in the order of organoid id and channel. 
            if debug_viz:
                import seaborn as sns

                ch_colors = sns.color_palette('Set1', len(org_tracks_list))

                fig, ax = plt.subplots()
                ax.imshow(obj_vid[0], alpha=0.5) # just visualise the first frame is enough.
            
            for ind_ids_ccc in ind_ids_cc:
                org_id, org_ch = ind_ids_ccc.split('_')
                org_id = int(org_id)
                org_ch = int(org_ch)
                
                boundary_org = org_tracks_list[org_ch][org_id]
                print(boundary_org)

                # # this is the problem? 
                # # objectness score for deciding the dominant organoid. 
                # obj_score = [ np.logical_not(org_tra_frame[0]) for org_tra_frame in boundary_org]
                # obj_score = np.sum(obj_score)
                # obj_score_cc.append(obj_score)  # pick the longest track. 
                obj_score = objectness_score_tracks_time(boundary_org[:], 
                                                          obj_vid[...,org_ch], 
                                                          mean_func=obj_mean_func)
                obj_score_cc.append(obj_score)

                # stability score which is weighted on 2 factors. to determine which track. 
                nan_score = nan_stability_score_tracks(boundary_org)
                
                # # this should be a minimisation ..... ! 
                # # smooth_score = smoothness_score_tracks(boundary_org, 
                # #                                       mean_func=np.nanmean)
                # smooth_score = smoothness_score_tracks_iou(boundary_org, 
                #                                            mean_func=smoothness_mean_func)
                
                # tra_stable_scores_cc.append(weight_nan*nan_score+weight_smooth*smooth_score)
                tra_stable_scores_cc.append(nan_score)

            if debug_viz:
                ax.set_title( 'org: %s, stable: %s'  %(ind_ids_cc[np.argmax(obj_score_cc)], 
                                                       ind_ids_cc[np.argmax(obj_score_cc)]))
                plt.show()
            
            # stack all the scores.     
            obj_score_cc = np.hstack(obj_score_cc)
            tra_stable_scores_cc = np.hstack(tra_stable_scores_cc)

            print(obj_score_cc)
            print(tra_stable_scores_cc)
            print('===')

            # decide on the organoid and track (argmax)
            cliq_org_id_keep = ind_ids_cc[np.argmax(obj_score_cc)]
            cliq_track_id_keep = ind_ids_cc[np.argmax(obj_score_cc)]
            
            # save this out for processing. 
            assigned_cliq_track_ids.append([cliq_org_id_keep, cliq_track_id_keep])

        # 4. new org_tracks_list production based on the filtered information. 
        org_tracks_list_out = []

        for list_ii in range(len(org_tracks_list)):

            org_tracks_list_ii = org_tracks_list[list_ii]
            org_tracks_list_ii_out = []

            for org_ii in range(len(org_tracks_list_ii)):
                tra_int_id = str(org_ii)+'_'+str(list_ii) # create the string id lookup. 
                
                include_track = True

                for cliq_ii in range(len(cliq_ids)):
                    ind_ids_cc = ind_ids[cliq_ids[cliq_ii]] # gets the clique members in string form -> is this organoid part of a clique. 
                    
                    if tra_int_id in ind_ids_cc:
                        include_track = False # do not automatically include.  

                        # test is this the dominant organoid in the clique. 
                        cliq_organoid_assign, cliq_organoid_assign_track = assigned_cliq_track_ids[cliq_ii] # get the assignment information of the clique. 
                        
                        if tra_int_id == cliq_organoid_assign:
                            # if this is the dominant organoid then we add the designated track. 
                            org_id_tra_assign, org_ch_tra_assign = cliq_organoid_assign_track.split('_')
                            org_id_tra_assign = int(org_id_tra_assign); org_ch_tra_assign=int(org_ch_tra_assign)
                            
                            org_tracks_list_ii_out.append(org_tracks_list[org_ch_tra_assign][org_id_tra_assign])

                        # do nothing otherwise -> exclude this organoid basically. 
                
                if include_track: 
                    # directly include. 
                    org_tracks_list_ii_out.append(org_tracks_list_ii[org_ii])

            if len(org_tracks_list_ii_out) > 0:
                # stack the tracks. 
                org_tracks_list_ii_out = np.array(org_tracks_list_ii_out)
            
            org_tracks_list_out.append(org_tracks_list_ii_out)
    else:
        org_tracks_list_out = list(org_tracks_list)

    # cleaned tracks, in the same input format as a list of numpy tracks for each channel. 
    return org_tracks_list_out


def filter_nan_tracks( boundary ):
    
    """
    function removes all tracks that for the entire duration is only nan. 
    """
    import numpy as np 
    
    boundaries_out = []
    
    for ii in range(len(boundary)):
        
        tra = boundary[ii]
        tra_len = len(tra)
        n_nans = 0
        
        for tra_tt in tra:
            if np.isnan(tra_tt[0]):
                n_nans+=1
        if n_nans < tra_len:
            boundaries_out.append(tra) # append the whole track
            
    if len(boundaries_out) > 0:
        boundaries_out = np.array(boundaries_out)
        
    return boundaries_out




import numpy as np 

def montage_vol_proj(vol, proj_fn=np.max):
    r"""Given a 3D grayscale volumetric imageand a projection function such as np.max, generates projected views onto x-y, x-z, y-z and montages them into one view.  

    Parameters
    ----------
    vol : (M,N,L) numpy array
        input grayscale volumetric image
    proj_fn : Python function
        this is any suitable projection function that allows the call format, proj_fn(vol, axis=0) to generate a 2D projection image.
    
    Returns
    -------
    montage_img : (N+M, L+L) 2D numpy array
        montaged 2D image which puts [top left] view_12, [top right] view_02, [bottom left] view_01
    """
    mm, nn, ll = vol.shape

    proj12 = proj_fn(vol, axis=0)
    proj02 = proj_fn(vol, axis=1)
    proj01 = proj_fn(vol, axis=2)

    vol_new = np.zeros((proj12.shape[0]+proj01.shape[0], proj12.shape[1]+proj02.shape[1]))
    
    vol_new[:proj12.shape[0], :proj12.shape[1]] = proj12.copy()
    vol_new[:proj02.shape[0], proj12.shape[1]:] = proj02.copy()
    vol_new[proj12.shape[0]:, :proj01.shape[1]] = proj01.copy()
    
    return vol_new





def imshowpair(im1,im2, ax=None):

    r""" Combines two potentially different sized grayscale images into one image. 

    Parameters
    ----------
    im1 : (M1,N1) grayscale numpy array
        input image 1 
    im2 :  (M2,N2) grayscale numpy array 
        input image 2 
    ax : Matplotlib axes object
        optional axis to display the image
    
    Returns
    -------
    img : (M_max,N_max,3) numpy array
        the combined image of im1 and im2, where M_max=max(M_1, M_2), and N_max=(N_1,N_2) are the largest row and column dimensions of im1, im2. The images are automatically centered in the output. 
    """

    from skimage.exposure import rescale_intensity
    import pylab as plt
    import numpy as np 
    
    dtype = type(im1.ravel()[0]) # check the data type
    shape1 = np.array(im1.shape)
    shape2 = np.array(im2.shape)
    
    img_shape = np.max(np.vstack([shape1,shape2]), axis=0)
    img = np.zeros((img_shape[0], img_shape[1], 3), dtype=dtype)
    
    offset1x = (img_shape[0] - shape1[0]) // 2; offset1y = (img_shape[1]-shape1[1]) // 2;
    offset2x = (img_shape[0] - shape2[0]) // 2; offset2y = (img_shape[1]-shape2[1]) // 2;
    
    # display centered images. 
    img[offset1x:offset1x+shape1[0],offset1y:offset1y+shape1[1],0] = im1
    img[offset2x:offset2x+shape2[0],offset2y:offset2y+shape2[1],1] = im2 
    
    if ax is not None:
        ax.imshow(rescale_intensity(img))
    
    return img
    
def checkerboard_imgs(im1, im2, grid=(10,10)):
    
    r""" generate a checkerboard montage of two 2D images, useful for checking misalignment when registering two images. 
    
    Parameters
    ----------
    im1 : (M,N) grayscale or (M,N,3) color image as numpy array
        input image 1
    im2 : (M,N) grayscale or (M,N,3) color image as numpy array
        input image 2
    grid : (int,int) tuple
        specifies the number of rows and number of columns in the gridding of the checkerboard. 

    Returns
    -------
    im : (M,N,3) numpy array
        the checkerboard combined image of im1 and im2. 
    """

    import numpy as np
    import pylab as plt

    # im1, im2 are grayscale or rgb images only. 
    if len(im1.shape) == 2:
        # grayscale image.
        rows, cols = im1.shape
    if len(im1.shape) == 3:
        # rgb image.
        rows, cols, _ = im1.shape
        
    # set up return image
    im = np.zeros((rows, cols, 3))
    
    # create the checkerboard mask.
    check_rows = np.linspace(0, rows, grid[0]+1).astype(np.int32)
    check_cols = np.linspace(0, cols, grid[1]+1).astype(np.int32)
    checkerboard = np.zeros((rows,cols))
    
    for i in range(grid[0]):
        
        if np.mod(i,2) == 0:
            even = 0
        else:
            even = 1
            
        for j in range(grid[1]):
            r = [check_rows[i], check_rows[i+1]]
            c = [check_cols[j], check_cols[j+1]]
            checkerboard[r[0]:r[1], c[0]:c[1]] = even 
            
            if even == 0:
                even = 1
            else:
                even = 0
                    
    ones = np.array(np.where(checkerboard==1)).T
    zeros = np.array(np.where(checkerboard==0)).T
        
    if len(im1.shape) == 2:
        # grayscale image.
        im[ones[:,0], ones[:,1], 0] = im1[ones[:,0], ones[:,1]]
        im[zeros[:,0], zeros[:,1], 1] = im2[zeros[:,0], zeros[:,1]]
        
    if len(im1.shape) == 3:
        # rgb image.
        im[ones[:,0], ones[:,1], :] = im1[ones[:,0], ones[:,1], :]
        im[zeros[:,0], zeros[:,1], :] = im2[zeros[:,0], zeros[:,1], :]
    
    return im 





def viz_grid(im_array, shape=None, cmap='gray', figsize=(10,10)):

    r""" given an array of images, plots them montage style in matplotlib 

    Parameters
    ----------
    im_array : n_img x n_rows x n_cols (x 3) grayscale or color numpy array
        array of n_img images to display
    shape :  (i,j) integer tuple 
        specifys the number of images to display row- and column- wise, if not set, will try np.sqrt, and make as square a panel as possible.
    cmap : string
        optional specification of a named matplotlib colormap to apply to view the image. 
    figsize : (int,int) tuple
        specify the figsize in matplotlib plot, implicitly controlling viewing resolution

    Returns
    -------
    (fig, ax): figure and ax handles 
        returned for convenience for downstream accessing of plot elements.
    """
    
    import pylab as plt 
    import numpy as np 
    
    n_imgs = len(im_array)
    
    if shape is not None:
        nrows, ncols = shape
    else:
        nrows = int(np.ceil(np.sqrt(n_imgs)))
        ncols = nrows
    
    color=True
    if len(im_array.shape) == 3:
        color=False
    if len(im_array.shape) == 4 and im_array.shape[-1]==1:
        color=False
    
    fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
    
    for i in range(nrows):
        for j in range(ncols):
            im_index = i*ncols + j 
            if im_index<n_imgs:
                if color:
                    ax[i,j].imshow(im_array[im_index])
                else:
                    ax[i,j].imshow(im_array[im_index], cmap=cmap)
            # switch off all gridlines. 
            ax[i,j].axis('off')
            ax[i,j].grid('off')
           
    return (fig, ax) 
    
    




"""

This module implements various different registration algorithms

"""
    
# these that are commented out should now be incorporated into individual functions below 
# from ..Utility_Functions import file_io as fio
# from ..Visualisation import imshowpair as imshowpair
# from ..Geometry import geometry as geom
# from transforms3d.affines import decompose44, compose
import numpy as np 

def matlab_affine_register(fixed_file, moving_file, save_file, reg_config, multiscale=False):
    r""" Use Python-Matlab interface to conduct affine registration. The transformed image is saved to file to bypass speed bottleneck in transferring Matlab to Python

    Parameters
    ----------
    fixed_file : filepath
        the filepath of the image to be used as reference for registration.
    moving_file : filepath
        the filepath of the image to be registered to the fixed image. 
    save_file : filepath
        the output filepath to save the registered image of the moving_file
    reg_config : dict
        Python dict specifying the parameters for running the registration. See :func:`unwrap3D.Parameters.params.affine_register_matlab` for the parameters required to be set. 
    multiscale : bool
        if True, do multiscale registration given by the ``register3D_intensity_multiscale`` script, otherwise run the ``register3D_rigid_faster`` script.

    Returns
    -------
    transform : 4x4 array
        the forward homogeneous transformation matrix to map fixed to moving. 
    im1 : array
        return the fixed image as numpy array from fixed_file if ``reg_config['return_img']!=1``. if ``reg_config['return_img']==1``, the transform is directly applied within Matlab and the result saved to save_file
    im2_ : array
        return the transformed moving image after applying the transform as a numpy array if ``reg_config['return_img']!=1``. if ``reg_config['return_img']==1``, the transform is directly applied within Matlab and the result saved to save_file
    """
    import matlab.engine
    import scipy.io as spio 
    import skimage.io as skio 
    from transforms3d.affines import decompose44, compose
    from ..Geometry import geometry as geom

    eng = matlab.engine.start_matlab() 
    
    if reg_config['view'] is not None:

        # use initialisation. 
        initial_tf = reg_config['initial_tf']

        # decompose the initialisation to comply with Matlab checks. # might be better ways to do this?
        T,R,Z,S = decompose44(initial_tf) # S is shear!
        
        if reg_config['type'] == 'similarity':
            # initialise with just rigid instead -> checks on this normally are too stringent.
            affine = compose(T, R, np.ones(3), np.zeros(3)) # the similarity criteria too difficult for matlab 
        if reg_config['type'] == 'rigid':
            affine = compose(T, R, np.ones(3), np.zeros(3))
        if reg_config['type'] == 'translation':
            affine = compose(T, np.eye(3), np.ones(3), np.zeros(3))
        if reg_config['type'] == 'affine':
            affine = initial_tf.copy()
        
        # save the tform as temporary for matlab to read. 
        spio.savemat('tform.mat', {'tform':affine}) # transpose for matlab 
        
        if multiscale == False:
            transform = eng.register3D_rigid_faster(str(fixed_file), str(moving_file), str(save_file), 
                                                    'tform.mat', 1, reg_config['downsample'], 
                                                    reg_config['modality'], reg_config['max_iter'], 
                                                    reg_config['type'], 
                                                    reg_config['return_img'], 
                                                    nargout=1) 
        else:
            transform = eng.register3D_intensity_multiscale(str(fixed_file), str(moving_file), str(save_file), 
                                                    'tform.mat', 1, reg_config['downsample'], 
                                                    reg_config['modality'], reg_config['max_iter'], 
                                                    reg_config['type'], 
                                                    reg_config['return_img'], 
                                                    nargout=1)
    else:
        if multiscale == False:
            transform = eng.register3D_rigid_faster(str(fixed_file), str(moving_file), str(save_file), 
                                                    'tform.mat', 0, reg_config['downsample'], 
                                                    reg_config['modality'], reg_config['max_iter'], 
                                                    reg_config['type'], 
                                                    reg_config['return_img'], 
                                                    nargout=1)    
        else:
            # print('multiscale')
            # convert to matlab arrays. 
            levels = matlab.double(reg_config['downsample'])
            warps = matlab.double(reg_config['max_iter'])
            transform = eng.register3D_intensity_multiscale(str(fixed_file), str(moving_file), str(save_file), 
                                                    'tform.mat', 0, levels, 
                                                    reg_config['modality'], warps, 
                                                    reg_config['type'], 
                                                    reg_config['return_img'], 
                                                    nargout=1)
    transform = np.asarray(transform)
    
    if reg_config['return_img']!= 1:
        # then the transform is applied within python 
        im1 = skio.imread(fixed_file)
        im2 = skio.imread(moving_file)
        im2_ = np.uint8(geom.apply_affine_tform(im2.transpose(2,1,0), 
                        np.linalg.inv(transform), 
                        np.array(im1.shape)[[2,1,0]]))
        im2_ = im2_.transpose(2,1,0) # flip back
    
        return transform, im1, im2_
    else:
        return transform 
        

# add multiscale capability 
def matlab_group_register_batch(dataset_files, ref_file, in_folder, out_folder, reg_config, reset_ref_steps=0, multiscale=True, debug=False):
    
    r""" Temporal affine registration of an entire video using Matlab call.

    Parameters
    ----------
    dataset_files : list of files
        list of img files arranged in increasing temporal order
    ref_file : str
        filepath of the initial reference file for registration. 
    in_folder : str
        specifies the parent folder of the dataset_files. This string is replaced with the out_folder string in order to generate the save paths for the registered timepoints.
    reg_config : dict
        Python dict specifying the parameters for running the registration. See :func:`unwrap3D.Parameters.params.affine_register_matlab` for the parameters required to be set. 
    reset_ref_steps : int
        if reset_ref_steps > 0, uses every nth registered file as the reference. default of 0 = fixed reference specified by ref_file, whilst 1 = sequential    
    multiscale : bool
        if True, do multiscale registration given by the ``register3D_intensity_multiscale`` script, otherwise run the ``register3D_rigid_faster`` script.
    debug : bool
        if True, will plot for every timepoint the 3 mid-slices in x-y, y-z, x-z to qualitatively check registration. Highly recommended for checking parameters.  
    
    Returns
    -------
    tforms : array
        the compiled 4x4 transformation matrices between all successive timepoints
    tformfile : str
        the filepath to the saved transforms 

    """
    import matlab.engine
    import scipy.io as spio 
    import os
    import shutil
    import pylab as plt 
    from tqdm import tqdm
    import skimage.io as skio 
    from ..Visualisation import imshowpair as imshowpair
    from ..Geometry import geometry as geom
    from ..Utility_Functions import file_io as fio

    # start matlab engine. 
    eng = matlab.engine.start_matlab() 
    fio.mkdir(out_folder) # check output folder exists. 
    
    tforms = []
    ref_files = []
    
    # difference is now fixed_file is always the same one. 
    fixed_file = ref_file
    all_save_files = np.hstack([f.replace(in_folder, out_folder) for f in dataset_files])
    
    for i in tqdm(range(len(dataset_files))):
    
        moving_file = dataset_files[i]
        save_file = all_save_files[i]
        
        if multiscale:
            # print('multiscale')
            levels = matlab.double(reg_config['downsample'])
            warps = matlab.double(reg_config['max_iter'])
            transform = eng.register3D_intensity_multiscale(str(fixed_file), str(moving_file), str(save_file), 
                                                    'tform.mat', 0, levels, 
                                                    reg_config['modality'], warps, 
                                                    reg_config['type'], 
                                                    reg_config['return_img'], 
                                                    nargout=1)
        else:
            transform = eng.register3D_rigid_faster(str(fixed_file), str(moving_file), str(save_file), 
                                            'tform.mat', 0, reg_config['downsample'], 
                                            reg_config['modality'], reg_config['max_iter'], 
                                            reg_config['type'], 
                                            reg_config['return_img'], 
                                            nargout=1)        
        
        transform = np.asarray(transform) # (z,y,x) 
        tforms.append(transform)
        ref_files.append(fixed_file) # which is the reference used. 

#        if reg_config['return_img'] != 1: # this is too slow and disabled. 
        im1 = skio.imread(fixed_file)
        im2 = skio.imread(moving_file)
        
        im2_ = np.uint8(geom.apply_affine_tform(im2.transpose(2,1,0), np.linalg.inv(transform), sampling_grid_shape=np.array(im1.shape)[[2,1,0]]))
        im2_ = im2_.transpose(2,1,0) # flip back

        # fio.save_multipage_tiff(im2_, save_file)
        skio.imsave(save_file, im2_)

        if debug:

            # visualize all three possible cross sections for reference and checking.
            fig, ax = plt.subplots(nrows=1, ncols=2)
            imshowpair(ax[0], im1[im1.shape[0]//2], im2[im2.shape[0]//2]) 
            imshowpair(ax[1], im1[im1.shape[0]//2], im2_[im2_.shape[0]//2])
            
            fig, ax = plt.subplots(nrows=1, ncols=2)
            imshowpair(ax[0], im1[:,im1.shape[1]//2], im2[:,im2.shape[1]//2])
            imshowpair(ax[1], im1[:,im1.shape[1]//2], im2_[:,im2_.shape[1]//2])

            fig, ax = plt.subplots(nrows=1, ncols=2)
            imshowpair(ax[0], im1[:,:,im1.shape[2]//2], im2[:,:,im2.shape[2]//2])
            imshowpair(ax[1], im1[:,:,im1.shape[2]//2], im2_[:,:,im2_.shape[2]//2])

            plt.show()
        
        if reset_ref_steps > 0:
            # change the ref file.
            if np.mod(i+1, reset_ref_steps) == 0: 
                fixed_file = save_file # change to the file you just saved. 
        
    # save out tforms into a .mat file.
    tformfile = os.path.join(out_folder, 'tforms-matlab.mat')
    tforms = np.array(tforms)
    ref_files = np.hstack(ref_files)
    
    spio.savemat(tformfile, {'tforms':tforms,
                             'ref_files':ref_files,
                             'in_files':dataset_files,
                             'out_files':all_save_files,
                             'reg_config':reg_config})
            
    return tforms, tformfile
    
        
def COM_2d(im1, im2):
    r""" Compute the center of mass between two images, im1 and im2 based solely on binary Otsu Thresholding. This can be used to compute the rough translational displacement.
    
    Parameters
    ----------
    im1 : array
        n-dimension image 1
    im2 : array
        n-dimension image 2
    
    Returns
    -------
    com1 : array
        n-dimension center of mass coordinates of image 1
    com2 : 
        n-dimension center of mass coordinates of image 2

    """
    from scipy.ndimage.measurements import center_of_mass
    from scipy.ndimage.morphology import binary_fill_holes
    from skimage.filters import threshold_otsu
    
    mask1 = binary_fill_holes(im1>=threshold_otsu(im1))
    mask2 = binary_fill_holes(im2>=threshold_otsu(im2))

    com1 = center_of_mass(mask1)
    com2 = center_of_mass(mask2)

    return com1, com2 
    
    
# function to join two volume stacks 
def simple_join(stack1, stack2, cut_off=None, blend=True, offset=10, weights=[0.7,0.3]):
    r""" Joins two volumetric images, stack1 and stack2 in the first dimension by direct joining or linear weighting constrained to +/- offset to the desired ``cut_off`` slice number.
    
    The constrained linear blending within the interval :math:`[x_c-\Delta x, x_c + \Delta x]` is given by 

    .. math::
        I = 
        \Biggl \lbrace 
        { 
        w_2\cdot I_1 + w_1\cdot I_2 ,\text{ if } 
          {x\in [x_c, x_c + \Delta x]}
        \atop 
        w_1\cdot I_1 + w_2\cdot I_2, \text{ if } 
          {x\in [x_c-\Delta x, x_c]}
        }
        
    where :math:`x_c` denotes the ``cut_off`` point. The weight reversal enables asymmetric weights to be define once but used twice. 

    Parameters
    ----------
    stack1 : (MxNxL) array
        volume image whose intensities will dominant in the first dimension when < ``cut_off``
    stack2 : (MxNxL) array
        volume image whose intensities will dominant in the first dimension when > ``cut_off``
    cut_off : int
        the slice number of the first dimension where the transition from stack2 to stack1 will occur
    blend : bool
        if blend==True, the join occurs directly at the cut_off point with simple array copying and there is no blending. If True then the linear blending as described occurs around the cut_off point. 
    offset : int
        the +/- slice numbers around the cut_off to linearly blend the intensities of stack1 and stack2 for a smoother effect.
    weights : 2 vector array
        :math:`w_1=` ``weights[0]`` and :math:`w_2=` ``weights[1]``. The sum of weights should be 1. 

    Returns
    -------
    the blended image of the same image dimension  

    """
    if cut_off is None:
        cut_off = len(stack1) // 2

    if blend:
        combined_stack = np.zeros_like(stack1)
        combined_stack[:cut_off-offset] = stack1[:cut_off-offset]
        combined_stack[cut_off-offset:cut_off] = weights[0]*stack1[cut_off-offset:cut_off]+weights[1]*stack2[cut_off-offset:cut_off]
        combined_stack[cut_off:cut_off+offset] = weights[1]*stack1[cut_off:cut_off+offset]+weights[0]*stack2[cut_off:cut_off+offset]
        combined_stack[cut_off+offset:] = stack2[cut_off+offset:]         
    else:
        combined_stack = np.zeros_like(stack1)
        combined_stack[:cut_off] = stack1[:cut_off] 
        combined_stack[cut_off:] = stack2[cut_off:] 
    
    return combined_stack


def sigmoid_join(stack1, stack2, cut_off=None, gradient=200, shape=1, debug=False):
    r""" Joins two volumetric images, stack1 and stack2 in the first dimension using a sigmoid type blending function. Both stacks are assumed to be the same size. 
    
    The sigmoid has the following mathematical expression 

    .. math::
        w_2 &= \frac{1}{\left( 1+e^{-\text{grad} (x - x_c)} \right)^{1/\text{shape}}} \\
        w_1 &= 1 - w_2
        
    where :math:`x_c` denoting the ``cut_off``.
    
    The blended image is then the weighted sum 

    .. math::
        I_{blend} = w_1\cdot I_1 + w_2\cdot I_2

    see https://en.wikipedia.org/wiki/Generalised_logistic_function

    Parameters
    ----------
    stack1 : (MxNxL) array
        volume image whose intensities will dominant in the first dimension when < ``cut_off``
    stack2 : (MxNxL) array
        volume image whose intensities will dominant in the first dimension when > ``cut_off``
    cut_off : int
        the slice number of the first dimension where the transition from stack2 to stack1 will occur
    gradient : scalar
        controls the sharpness of the transition between the two stacks 
    shape : scalar
        controls the shape of the sigmoid, in particular introduces asymmetry in the shape
    debug : bool
        if True, will use Matplotlib to plot the computed sigmoid weights for checking.   

    Returns
    -------
    the blended image of the same image dimension  

    """
    if cut_off is None:
        cut_off = len(stack1) // 2

    def generalised_sigmoid( stack1, cut_off=cut_off, shape=shape, grad=gradient):
        
        x = np.arange(0,stack1.shape[0])
        weights2 = 1./((1+np.exp(-grad*(x - cut_off)))**(1./shape))
        weights1 = 1. - weights2
        
        return weights1, weights2
    
    weights1, weights2 = generalised_sigmoid( stack1, cut_off=cut_off, shape=shape, grad=gradient)
    if debug:
        import pylab as plt 
        plt.figure()
        plt.plot(weights1, label='1')
        plt.plot(weights2, label='2')
        plt.legend()
        plt.show()
        
    return stack1*weights1[:,None,None] + stack2*weights2[:,None,None]


def nonregister_3D_demons_matlab(infile1, infile2, savefile, savetransformfile, reg_config):
    r""" This function uses Matlab's imregdemons function to run Demon's registration. The result is directly saved to disk.
    
    Parameters
    ----------
    infile1 : str
        filepath of the reference static image
    infile2 : str
        filepath of the moving image to register to the fixed image
    savefile : str
        filepath to save the transformed moving image to 
    savetransformfile : str
        filepath to save the displacement field to 
    reg_config : dict
        parameters to pass to imregdemons. see :func:`unwrap3D.Parameters.params.demons_register_matlab` 

    Returns
    -------
    return_val : int
        return value of 1 if Matlab execution was successful and completed
    """
    import matlab.engine
    eng = matlab.engine.start_matlab()
    
    print(reg_config['level'], reg_config['warps'])
    
#    alpha = matlab.single(reg_config['alpha'])
#    level = matlab.double(reg_config['level'])
    level = float(reg_config['level'])
    warps = matlab.double(reg_config['warps'])
    smoothing = float(reg_config['alpha'])
    
    # add in additional options for modifying. 
    return_val = eng.nonrigid_register3D_demons(str(infile1), str(infile2), str(savefile), 
                                        str(savetransformfile),
                                        level, warps, smoothing)
        
    return return_val  


def warp_3D_demons_tfm(infile, savefile, transformfile, downsample, direction=1):
    r""" This function warps the input image file according to the deformation field specified in transformfile and saves the result in savefile. 
    If direction == 1 warp in the same direction else if direction == -1 in the reverse direction.
    
    Parameters
    ----------
    infile : str 
        filepath of the image to transform 
    savefile : str
        filepath of transformed result
    transformfile : str
        filepath to the displacement field transform 
    downsample : int
        this parameter specifies the highest level of downsampling the displacement field came from i.e. if the displacement field is the same size as the image then downsample=1. If the displacement field was generated at a scale 4x smaller than the input image then specify downsample=4 here.
        The displacment field will be rescaled by this factor before applying to the input image.
    direction : 1 or -1 
        if 1, warps the image in the same direction of the diplacement field. If -1 warps the image in the reverse direction of the displacement field

    Returns
    -------
    return_val : int
        if 1, the process completed successfully in Matlab 
    """
    import matlab.engine
    eng = matlab.engine.start_matlab()

    return_val = eng.warp_3D_demons(str(infile), str(savefile), str(transformfile), int(downsample), direction)

    return return_val


def warp_3D_demons_matlab_tform_scipy(im, tformfile, direction='F', pad=20):
    r""" Python-based warping of Matlab imregdemons displacement fields adjusting for the difference in coordinate convention.

    Parameters
    ----------
    im : (MxNxL) array
        input 3D volumetric image to warp
    tformfile : str 
        filepath to the displacement field transform 
    direction : 'F' or 'B'
        if 'F' warps im in the forwards direction specified by the displacement field or if 'B' warps im in the backwards direction.
    pad : int
        optional isotropic padding applied to every image dimension before warping

    Returns
    -------
    im_interp : (MxNxL) array
        the warped 3D volumetric image

    """
    import numpy as np 
    from scipy.ndimage import map_coordinates
    from ..Utility_Functions import file_io as fio
    
    dx,dy,dz = fio.read_demons_matlab_tform(tformfile, im.shape)
    # print(dx.max(), dx.min())
    im_ = im.copy()
    if pad is not None:
        im_ = np.pad(im_, [[pad,pad] for i in range(len(im_.shape))], mode='constant') # universally pad the volume by the edge values. 
        dx = np.pad(dx, [[pad,pad] for i in range(len(dx.shape))], mode='constant') # universally pad the volume by the edge values. 
        dy = np.pad(dy, [[pad,pad] for i in range(len(dy.shape))], mode='constant') # universally pad the volume by the edge values. 
        dz = np.pad(dz, [[pad,pad] for i in range(len(dz.shape))], mode='constant') # universally pad the volume by the edge values. 

    im_interp = warp_3D_displacements_xyz(im_, dx, dy, dz, direction=direction)
    
    return im_interp


def warp_3D_displacements_xyz(im, dx, dy, dz, direction='F', ret_xyz=False, mode='nearest'):
    r""" Warps an image with a 3D displacement field specified by arrays, optionally returning the corresponding coordinates where the image intensities were interpolated from in the image

    Parameters
    ----------
    im : (MxNxL) array
        the input 3D volume image
    dx : (MxNxL) array
        the displacements in x-direction (first image axis)
    dy : (MxNxL) array
        the displacements in y-direction (second image axis)
    dz : (MxNxL) array
        the displacements in z-direction (third image axis)
    direction : 'F' or 'B'
        if 'F' warp image in the same direction as specified by (dx,dy,dz) or if 'B' warp image in the opposite direction of (-dx,-dy,-dz)
    ret_xyz : bool
        if True, return as a second output the x-, y- and z- coordinates in ``im`` where ``im_interp`` intensities were interpolated at
    mode : str
         one of the boundary modes specified as ``mode`` argument in `scipy.map_coordinates <https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.map_coordinates.html>`_

    Returns
    -------
    im_interp : (MxNxL) array
        the transformed image 
    [XX, YY, ZZ] : list of arrays
        the x-, y- and z- coordinates of same array shape as im where ``im_interp`` intensities were taken from ``im`` 

    """    
    from scipy.ndimage import map_coordinates
    import numpy as np 
    
    XX, YY, ZZ = np.indices(im.shape) # set up the interpolation grid.
    
    if direction == 'F':
        XX = XX + dx
        YY = XX + dy 
        ZZ = ZZ + dz
    if direction == 'B':
        XX = XX - dx
        YY = YY - dy
        ZZ = ZZ - dz
    
    im_interp = map_coordinates(im, 
                                np.vstack([(XX).ravel().astype(np.float32), 
                                           (YY).ravel().astype(np.float32), 
                                           (ZZ).ravel().astype(np.float32)]), prefilter=False, order=1, mode=mode) # note the mode here refers to the boundary handling.
    im_interp = im_interp.reshape(im.shape)
    
    if ret_xyz:
        return im_interp, [XX,YY,ZZ]
    else:
        return im_interp


def warp_3D_displacements_xyz_pts(pts3D, dxyz, direction='F'):
    r""" Convenience function for warping forward or backward a 3D point cloud given the displacement field 

    If direction == 'F', the new points, :math:`(x',y',z')` are

    .. math::
        (x',y',z') = (x+dx, y+dy, z+dz)

    If direction == 'B', the new points, :math:`(x',y',z')` are

    .. math::
        (x',y',z') = (x-dx, y-dy, z-dz)

    Parameters
    ----------
    pts3D : array
        n-dimensional array with last dimension of size 3 for (x,y,z)
    dxyz : array 
        n-dimensional array with last dimension of size 3 for (dx,dy,dz)
    direction : 'F' or 'B'
        if 'F' warp points in the same direction as specified by (dx,dy,dz) or if 'B' warp image in the opposite direction of (-dx,-dy,-dz)
    
    Returns
    -------
    pts3D_warp : array
        n-dimensional array with last dimension of size 3 for the new (x,y,z) coordinates after warping

    """    
    if direction == 'F':
        pts3D_warp = pts3D + dxyz
    if direction == 'B':
        pts3D_warp = pts3D - dxyz
    
    return pts3D_warp

def warp_3D_transforms_xyz(im, tmatrix, direction='F', ret_xyz=False):
    r""" Warps a volume image with the given affine transformation matrix forward or reverse with trilinear interpolation, returning if specified the corresponding (x,y,z) grid points

    Parameters
    ----------
    im : (MxNxL) array
        the input 3D volume image
    tmatrix : (4x4) array 
        the homogeneous affine 3D transformation matrix 
    direction : 'F' or 'B'
        if 'F' warp image with tmatrix or if 'B' warp image with the matrix inverse of tmatrix
    ret_xyz : bool
        if True, return as a second output the xyz- coordinates in ``im`` where ``im_interp`` intensities were interpolated at
    
    Returns
    -------
    im_interp : (MxNxL) array
        the transformed image 
    xyz_ : list of arrays
        the xyz- coordinates of same array shape as im where ``im_interp`` intensities were taken from ``im`` 

    """    
    import numpy as np 
    from scipy.ndimage import map_coordinates

    XX, YY, ZZ = np.indices(im.shape) # set up the interpolation grid.
    
    xyz = np.vstack([(XX).ravel().astype(np.float32), 
                     (YY).ravel().astype(np.float32), 
                     (ZZ).ravel().astype(np.float32),
                     np.ones(len(ZZ.ravel()), dtype=np.float32)])
    
    if direction == 'F':
        xyz_ = tmatrix.dot(xyz)
    if direction == 'B':
        xyz_ = (np.linalg.inv(tmatrix)).dot(xyz)
    
    im_interp = map_coordinates(im, 
                                xyz_[:3], prefilter=False, order=1, mode='nearest')
    im_interp = im_interp.reshape(im.shape)

    # return also the coordinates that it maps to 
    if ret_xyz:
        return im_interp, xyz_
    else:
        return im_interp

def warp_3D_transforms_xyz_pts(pts3D, tmatrix, direction='F'):
    r""" Convenience function for warping forward or backward a 3D point cloud given an affine transform matrix 

    Parameters
    ----------
    pts3D : array
        n-dimensional array with last dimension of size 3 for (x,y,z)
    tmatrix : (4x4) array 
        the homogeneous affine 3D transformation matrix 
    direction : 'F' or 'B'
        if 'F' warp points with ``tmatrix`` or if 'B' warp points with the reverse transformation given by``np.linalg.inv(tmatrix)``, the matrix inverse
    
    Returns
    -------
    pts3D_warp : array
        n-dimensional array with last dimension of size 3 for the new (x,y,z) coordinates after warping

    """    
    import numpy as np 
    # first make homogeneous coordinates.
    xyz = np.vstack([(pts3D[...,0]).ravel().astype(np.float32), 
                     (pts3D[...,1]).ravel().astype(np.float32), 
                     (pts3D[...,2]).ravel().astype(np.float32),
                     np.ones(len(pts3D[...,0].ravel()), dtype=np.float32)])
    
    if direction == 'F':
        xyz_ = tmatrix.dot(xyz)
    if direction == 'B':
        xyz_ = (np.linalg.inv(tmatrix)).dot(xyz)
    
    pts3D_warp = (xyz_[:3].T).reshape(pts3D.shape)

    return pts3D_warp


def warp_3D_transforms_xyz_similarity(im, translation=[0,0,0], rotation=np.eye(3), zoom=[1,1,1], shear=[0,0,0], center_tform=True, direction='F', ret_xyz=False, pad=50):
    r""" Warps a volume image forward or reverse  where the transformation matrix is explicitly given by the desired translation, rotation, zoom and shear matrices, returning if specified the corresponding (x,y,z) grid points and optional padding

    Parameters
    ----------
    im : (MxNxL) array
        the input 3D volume image
    translation : (3,) array 
        the global (dx,dy,dz) translation vector
    rotation : (3x3) array 
        the desired rotations given as a rotation matrix
    zoom : (3,) array
        the independent scaling factor in x-, y-, z- directions 
    shear : (3,) array
        the shearing factor in x-, y-, z- directions
    center_tform : bool
        if true, the transformation will be applied preserving ``im_center`` 
    direction : 'F' or 'B'
        if 'F' warp points with the specified transformation or if 'B' warp points by the reverse transformation
    ret_xyz : bool
        if True, return as a second output the xyz- coordinates in ``im`` where ``im_interp`` intensities were interpolated at
    pad : int 
        if not None, the input image is prepadded with the same number of pixels given by ``pad`` in all directions before transforming 
    
    Returns
    -------
    im_interp : (MxNxL) array
        the transformed image 
    xyz_ : list of arrays
        the xyz- coordinates of same array shape as im where ``im_interp`` intensities were taken from ``im`` 

    """    
    """
    This function is mainly to test how to combine the coordinates with transforms. 

    pad enables an optional padding of the image. 
    """
    from scipy.ndimage import map_coordinates
    from transforms3d.affines import compose
    import numpy as np 

    # compose the 4 x 4 homogeneous matrix. 
    tmatrix = compose(translation, rotation, zoom, shear)

    im_ = im.copy()

    if pad is not None:
        im_ = np.pad(im_, [[pad,pad] for i in range(len(im.shape))], mode='constant') # universally pad the volume by the edge values. 

    if center_tform:
        im_center = np.array(im_.shape)//2
        tmatrix[:-1,-1] = tmatrix[:-1,-1] + np.array(im_center)
        decenter = np.eye(4); decenter[:-1,-1] = -np.array(im_center)
        tmatrix = tmatrix.dot(decenter)

    XX, YY, ZZ = np.indices(im_.shape) # set up the interpolation grid.
    
    xyz = np.vstack([(XX).ravel().astype(np.float32), 
                     (YY).ravel().astype(np.float32), 
                     (ZZ).ravel().astype(np.float32),
                     np.ones(len(ZZ.ravel()), dtype=np.float32)])
    
    if direction == 'F':
        xyz_ = tmatrix.dot(xyz)
    if direction == 'B':
        xyz_ = (np.linalg.inv(tmatrix)).dot(xyz)
    
    # needs more memory % does this actually work? 
    im_interp = map_coordinates(im_, 
                                xyz_[:3], prefilter=False, order=1, mode='nearest')
    im_interp = im_interp.reshape(im_.shape)
    
    if ret_xyz:
        # additional return of the target coordinates. 
        return im_interp, xyz_
    else:
        return im_interp


def warp_3D_transforms_xyz_similarity_pts(pts3D, translation=[0,0,0], rotation=np.eye(3), zoom=[1,1,1], shear=[0,0,0], center_tform=True, im_center=None, direction='F'):
    r""" Convenience function for warping forward or backward a 3D point cloud where  the transformation matrix is explicitly given by the desired translation, rotation, zoom and shear matrices  

    Parameters
    ----------
    pts3D : array
        n-dimensional array with last dimension of size 3 for (x,y,z)
    translation : (3,) array 
        the global (dx,dy,dz) translation vector
    rotation : (3x3) array 
        the desired rotations given as a rotation matrix
    zoom : (3,) array
        the independent scaling factor in x-, y-, z- directions 
    shear : (3,) array
        the shearing factor in x-, y-, z- directions
    center_tform : bool
        if true, the transformation will be applied preserving ``im_center`` 
    im_center : (3,) array
        if None, the mean of ``pt3D`` in x-, y-, z- is taken as ``im_center`` and takes effect only if ``center_tform=True``
    direction : 'F' or 'B'
        if 'F' warp points with the specified transformation or if 'B' warp points by the reverse transformation
    
    Returns
    -------
    pts3D_warp : array
        n-dimensional array with last dimension of size 3 for the new (x,y,z) coordinates after warping

    """    
    from transforms3d.affines import compose 
    
    # compose the 4 x 4 homogeneous matrix. 
    tmatrix = compose(translation, rotation, zoom, shear)

    if center_tform:
        if im_center is None:
            im_center = np.hstack([np.nanmean(pts3D[...,ch]) for ch in pts3D.shape[-1]])
        # im_center = np.array(im.shape)//2
        tmatrix[:-1,-1] = tmatrix[:-1,-1] + np.array(im_center)
        decenter = np.eye(4); decenter[:-1,-1] = -np.array(im_center)
        tmatrix = tmatrix.dot(decenter)

    # first make homogeneous coordinates.
    xyz = np.vstack([(pts3D[...,0]).ravel().astype(np.float32), 
                     (pts3D[...,1]).ravel().astype(np.float32), 
                     (pts3D[...,2]).ravel().astype(np.float32),
                     np.ones(len(pts3D[...,0].ravel()), dtype=np.float32)])
    
    if direction == 'F':
        xyz_ = tmatrix.dot(xyz)
    if direction == 'B':
        xyz_ = (np.linalg.inv(tmatrix)).dot(xyz)
    
    pts3D_warp = (xyz_[:3].T).reshape(pts3D.shape)

    return pts3D_warp

    
def smooth_and_resample(image, shrink_factors, smoothing_sigmas):
    """ Utility function used in :func:`unwrap3D.Registration.registration.multiscale_demons` for generating multiscale image pyramids for registration based on the SimpleITK library

    see SimpleITK notebooks, https://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/Python_html/66_Registration_Demons.html

    Parameters
    ----------
    image : SimpleITK image
        The image we want to resample.
    shrink_factors : scalar or array
        Number(s) greater than one, such that the new image's size is original_size/shrink_factor.
    smoothing_sigma(s): scalar or array
        Sigma(s) for Gaussian smoothing, this is in physical units, not pixels.

    Returns
    -------
    image_resample : SimpleITK image
        Image which is a result of smoothing the input and then resampling it using the given sigma(s) and shrink factor(s).
    """
    import SimpleITK as sitk
    import numpy as np 

    if np.isscalar(shrink_factors):
        shrink_factors = [shrink_factors]*image.GetDimension()
    if np.isscalar(smoothing_sigmas):
        smoothing_sigmas = [smoothing_sigmas]*image.GetDimension()

    smoothed_image = sitk.SmoothingRecursiveGaussian(image, smoothing_sigmas)
    
    original_spacing = image.GetSpacing()
    original_size = image.GetSize()
    new_size = [int(sz/float(sf) + 0.5) for sf,sz in zip(shrink_factors,original_size)]
    new_spacing = [((original_sz-1)*original_spc)/(new_sz-1) 
                   for original_sz, original_spc, new_sz in zip(original_size, original_spacing, new_size)]

    image_resample = sitk.Resample(smoothed_image, new_size, sitk.Transform(), 
                         sitk.sitkLinear, image.GetOrigin(),
                         new_spacing, image.GetDirection(), 0.0, 
                         image.GetPixelID())
    
    return image_resample
    

def multiscale_demons(registration_algorithm,
                      fixed_image, moving_image, initial_transform = None, 
                      shrink_factors=None, smoothing_sigmas=None):
    r""" Run the given registration algorithm in a multiscale fashion. The original scale should not be given as input as the
    original images are implicitly incorporated as the base of the pyramid.
    
    We use the algorithm here to run demons registration in multiscale hence the name.

    see SimpleITK notebooks, https://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/Python_html/66_Registration_Demons.html

    Parameters
    ----------
    registration_algorithm : SimpleITK algorithm instance
        Any registration algorithm in `SimpleITK <https://simpleitk.readthedocs.io/en/master/filters.html>`_ that has an Execute(fixed_image, moving_image, displacement_field_image) method.
    fixed_image: SimpleITK image
        This is the reference image. Resulting transformation maps points from this image's spatial domain to the moving image spatial domain. 
    moving_image : SimpleITK image
        This is the image to register to the fixed image. Resulting transformation maps points from the fixed_image's spatial domain to this image's spatial domain.
    initial_transform : SimpleITK transform
        Any SimpleITK transform, used to initialize the displacement field.
    shrink_factors : list of lists or scalars 
        Shrink factors relative to the original image's size. When the list entry, shrink_factors[i], is a scalar the same factor is applied to all axes.
        When the list entry is a list, shrink_factors[i][j] is applied to axis j. This allows us to specify different shrink factors per axis. This is useful
        in the context of microscopy images where it is not uncommon to have unbalanced sampling such as a 512x512x8 image. In this case we would only want to 
        sample in the x,y axes and leave the z axis as is: [[[8,8,1],[4,4,1],[2,2,1]].
    smoothing_sigmas : list of lists or scalars  
        Amount of smoothing which is done prior to resmapling the image using the given shrink factor. These are in physical (image spacing) units.

    Returns
    -------
    SimpleITK.DisplacementFieldTransform

    """
    import SimpleITK as sitk
    import numpy as np 

    # Create image pyramid in a memory efficient manner using a generator function.
    # The whole pyramid never exists in memory, each level is created when iterating over
    # the generator.
    def image_pair_generator(fixed_image, moving_image, shrink_factors, smoothing_sigmas):
        end_level = 0
        start_level = 0
        if shrink_factors is not None:
            end_level = len(shrink_factors)
        for level in range(start_level, end_level):
            f_image = smooth_and_resample(fixed_image, shrink_factors[level], smoothing_sigmas[level])
            m_image = smooth_and_resample(moving_image, shrink_factors[level], smoothing_sigmas[level])
            yield(f_image, m_image)
        yield(fixed_image, moving_image)
    
    # Create initial displacement field at lowest resolution. 
    # Currently, the pixel type is required to be sitkVectorFloat64 because 
    # of a constraint imposed by the Demons filters.
    if shrink_factors is not None:
        original_size = fixed_image.GetSize()
        original_spacing = fixed_image.GetSpacing()
        s_factors =  [shrink_factors[0]]*len(original_size) if np.isscalar(shrink_factors[0]) else shrink_factors[0]
        df_size = [int(sz/float(sf) + 0.5) for sf,sz in zip(s_factors,original_size)]
        df_spacing = [((original_sz-1)*original_spc)/(new_sz-1) 
                      for original_sz, original_spc, new_sz in zip(original_size, original_spacing, df_size)]
    else:
        df_size = fixed_image.GetSize()
        df_spacing = fixed_image.GetSpacing()
 
    if initial_transform:
        initial_displacement_field = sitk.TransformToDisplacementField(initial_transform, 
                                                                       sitk.sitkVectorFloat64,
                                                                       df_size,
                                                                       fixed_image.GetOrigin(),
                                                                       df_spacing,
                                                                       fixed_image.GetDirection())
    else:
        initial_displacement_field = sitk.Image(df_size, sitk.sitkVectorFloat64, fixed_image.GetDimension())
        initial_displacement_field.SetSpacing(df_spacing)
        initial_displacement_field.SetOrigin(fixed_image.GetOrigin())
 
    # Run the registration.            
    # Start at the top of the pyramid and work our way down.    
    for f_image, m_image in image_pair_generator(fixed_image, moving_image, shrink_factors, smoothing_sigmas):
        initial_displacement_field = sitk.Resample (initial_displacement_field, f_image)
        initial_displacement_field = registration_algorithm.Execute(f_image, m_image, initial_displacement_field)
    return sitk.DisplacementFieldTransform(initial_displacement_field)


def SITK_multiscale_affine_registration(vol1, vol2, 
                                        imtype=16,
                                        p12=None,
                                        rescale_intensity=True,  
                                        centre_tfm_model='geometry', 
                                        tfm_type = 'rigid', 
                                        metric='Matte',  
                                        metric_numberOfHistogramBins=16,
                                        sampling_intensity_method='random',
                                        MetricSamplingPercentage=0.1,
                                        shrink_factors = [1], 
                                        smoothing_sigmas = [0],
                                        optimizer='gradient', 
                                        optimizer_params=None, 
                                        eps=1e-12):
    r""" Main function to affine register two input volumes using SimpleITK library in a multiscale manner. 
    
    Affine registration includes the following transformations

    - rigid (rotation + translation)
    - iso_similarity (isotropic scale + rigid)
    - aniso_similarity (anisotropic scale + rigid)
    - affine (skew + aniso_similarity)

    see SimpleITK notebooks, https://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/Python_html/61_Registration_Introduction_Continued.html

    Parameters
    ----------
    vol1 : (MxNxL) array
        reference volume given as a Numpy array
    vol2 : (MxNxL) array
        volume to register to vol2, given as a Numpy array
    imtype : int
        specifies the -bit of the image e.g. 8-bit, 16-bit etc. Used to normalize the intensity if ``rescale_intensity=False``
    p12 : (p1,p2) tuple
        if provided, vol1 and vol2 are constract enhanced using percentage intensity normalization 
    rescale_intensity : bool
        if True, does min-max scaling of image intensities in the two volumes independently before registering. This follows percentage intensity normalization if specified. 
    centre_tfm_model : str
        specifies the initial centering model. This is one of two options. For microscopy, we find 'geometry' has less jitter and is best. 
    tfm_type : str
        Type of affine transform from 4. The more complex, the longer the registration. 

        'rigid' :
            rotation + translation
        'iso_similarity' : 
            isotropic scale + rotation + translation 
        'aniso_similarity' :
             anisotropic scale + rotation + translation 
        'affine' : 
            skew + scale + rotation + translation 
    
    metric : str 
        This parameter is currently not used. Matte's mutual information ('Matte') is used by default as it offers the best metric for fluorescence imaging and multimodal registration.  
    metric_numberOfHistogramBins : int,
        The number of histogram bins to compute mutual information. Higher bins may allow finer registration but also adds more intensity noise. Smaller bins is more robust to intensity variations but may not be as discriminative.
    sampling_intensity_method : str
        This parameter is currently not used. 'random' is on by default. Evaluating the metric of registration is too expensive for the full image, thus we need to sample. Random is good since it avoids consistently evaluating the same grid points. 
    MetricSamplingPercentage : scalar,
        Number betwen 0-1 specifying the fraction of image voxels to evaluate the cost per registration iteration. Usually want to keep this fairly low for speed.  
    shrink_factors : list of lists or scalars 
        Shrink factors relative to the original image's size. When the list entry, shrink_factors[i], is a scalar the same factor is applied to all axes.
        When the list entry is a list, shrink_factors[i][j] is applied to axis j. This allows us to specify different shrink factors per axis. This is useful
        in the context of microscopy images where it is not uncommon to have unbalanced sampling such as a 512x512x8 image. In this case we would only want to 
        sample in the x,y axes and leave the z axis as is: [[[8,8,1],[4,4,1],[2,2,1]].
    smoothing_sigmas : list of lists or scalars  
        Amount of smoothing which is done prior to resmapling the image using the given shrink factor. These are in pixel units.
    optimizer : str 
        The optimization algorithm used to find the parameters. 

        'gradient' : 
            uses regular gradient descent. registration_method.SetOptimizerAsGradientDescent
        '1+1_evolutionary' : 
            uses 1+1 evolutionary which is default in Matlab. registration_method.SetOptimizerAsOnePlusOneEvolutionary

    optimizer_params : dict
        Python dictionary-like specification of the optimization parameters. See :func:`unwrap3D.Parameters.params.gradient_descent_affine_reg` for setting up gradient descent (Default) and 
        :func:`unwrap3D.Parameters.params.evolutionary_affine_reg` for setting up the evolutionary optimizer
    eps : scalar
        small number for numerical precision
    
    Returns
    -------
    SimpleITK.Transform

    """
    import SimpleITK as sitk

    def imadjust(vol, p1, p2): 
        import numpy as np 
        from skimage.exposure import rescale_intensity
        # this is based on contrast stretching and is used by many of the biological image processing algorithms.
        p1_, p2_ = np.percentile(vol, (p1,p2))
        vol_rescale = rescale_intensity(vol, in_range=(p1_,p2_))
        return vol_rescale

    import numpy as np 

    if p12 is None:
        im1_ = vol1.copy()
        im2_ = vol2.copy()
        # no contrast stretching 
        if rescale_intensity == False:
            im1_ = im1_ / (2**imtype - 1)
            im2_ = im2_ / (2**imtype - 1)
        else:
            im1_ = (im1_ - im1_.min()) / (im1_.max() - im1_.min() + eps)
            im2_ = (im2_ - im2_.min()) / (im2_.max() - im2_.min() + eps)
    else:
        # contrast stretching 
        im1_ = imadjust(vol1, p12[0], p12[1])
        im2_ = imadjust(vol2, p12[0], p12[1])

        if rescale_intensity == False:
            im1_ = im1_ / (2**imtype - 1)
            im2_ = im2_ / (2**imtype - 1)
        else:
            im1_ = (im1_ - im1_.min()) / (im1_.max() - im1_.min() + eps)
            im2_ = (im2_ - im2_.min()) / (im2_.max() - im2_.min() + eps)

    ### main analysis scripts. 
    v1 = sitk.GetImageFromArray(im1_, isVector=False)
    v2 = sitk.GetImageFromArray(im2_, isVector=False)
        
    # a) initial transform 
    # translation.
    if centre_tfm_model=='geometry':
        translation_mode = sitk.CenteredTransformInitializerFilter.GEOMETRY
    if centre_tfm_model=='moments':
        translation_mode = sitk.CenteredTransformInitializerFilter.MOMENTS

    # if tfm_type == 'translation':
    #     transform_mode = sitk.TranslationTransform(3)    
    # these are the built in transforms without additional modification. 
    
    if tfm_type == 'rigid':
        transform_mode = sitk.Euler3DTransform()
    elif tfm_type == 'iso_similarity':
        transform_mode = sitk.Similarity3DTransform()
    elif tfm_type == 'aniso_similarity':
        transform_mode = sitk.ScaleVersor3DTransform() # this version allows anisotropic scaling 
    elif tfm_type == 'affine': 
        # transform_mode = sitk.AffineTransform(3) # this is a generic that requires setting. 
        transform_mode = sitk.ScaleSkewVersor3DTransform() # this is anisotropic + skew
    else:
        print('error: no valid transform specfied')
    
    initial_transform = sitk.CenteredTransformInitializer(v1, 
                                                          v2, 
                                                          transform_mode, # this is where we change the transform
                                                          translation_mode)

    # a) Affine registration transform  
    # Select a different type of affine registration
    # multiscale rigid.
    registration_method = sitk.ImageRegistrationMethod()

    # Similarity metric settings.
    registration_method.SetMetricAsMattesMutualInformation(numberOfHistogramBins=metric_numberOfHistogramBins) 
    # key for making it fast. (subsampling, don't use all the pixels)
    if sampling_intensity_method == 'random':
        registration_method.SetMetricSamplingStrategy(registration_method.RANDOM)
    registration_method.SetMetricSamplingPercentage(MetricSamplingPercentage)
    registration_method.SetInterpolator(sitk.sitkLinear) # use this to help resampling the intensity at each iteration. 

    # Optimizer settings. # put these settings in a dedicated params file? 
    if optimizer == 'gradient':
        registration_method.SetOptimizerAsGradientDescent(learningRate=optimizer_params['learningRate'], 
                                                          numberOfIterations=optimizer_params['numberOfIterations'], # increase this.  
                                                          convergenceMinimumValue=optimizer_params['convergenceMinimumValue'], 
                                                          convergenceWindowSize=optimizer_params['convergenceWindowSize'],
                                                          estimateLearningRate=registration_method.EachIteration)
    if optimizer == '1+1_evolutionary':
        registration_method.SetOptimizerAsOnePlusOneEvolutionary(numberOfIterations=optimizer_params['numberOfIterations'], 
                                                                 epsilon=optimizer_params['epsilon'], 
                                                                 initialRadius= optimizer_params['initialRadius'],
                                                                 growthFactor= optimizer_params['growthFactor'],
                                                                 shrinkFactor= optimizer_params['shrinkFactor']
                                                                 )
    registration_method.SetOptimizerScalesFromIndexShift()

    # set the multiscale registration parameters here. 
    registration_method.SetShrinkFactorsPerLevel(shrinkFactors = shrink_factors) # use just the one scale. 
    registration_method.SetSmoothingSigmasPerLevel(smoothingSigmas=smoothing_sigmas) # don't filter. 
    registration_method.SetInitialTransform(initial_transform, inPlace=False)
    
    tfm = registration_method.Execute(sitk.Cast(v1, sitk.sitkFloat32), 
                                      sitk.Cast(v2, sitk.sitkFloat32))

    # if tfm_type == 'affine': 
    #     # refit with the above params. 
    #     tfm_affine = sitk.AffineTransform(3)
    #     tfm_affine.SetMatrix(tfm.GetMatrix())
    #     tfm_affine.SetTranslation(tfm.GetTranslation())
    #     tfm_affine.SetCenter(tfm.GetCenter())
    #     registration_method.SetInitialTransform(tfm_affine, inPlace=False)
    #     tfm = registration_method.Execute(sitk.Cast(v1, sitk.sitkFloat32), 
    #                                       sitk.Cast(v2, sitk.sitkFloat32))

    return tfm 
    

# to do: define helper script to wrap the main preprocessing steps for demons registration of two volumetric images.
def SITK_multiscale_demons_registration(vol1, vol2, 
                                        imtype=16,
                                        p12=(2,99.8),
                                        rescale_intensity=True,  
                                        centre_tfm_model='geometry', 
                                        demons_type='diffeomorphic', 
                                        n_iters = 25, 
                                        smooth_displacement_field = True,
                                        smooth_alpha=.8,
                                        shrink_factors = [2.,1.], 
                                        smoothing_sigmas = [1.,1.],
                                        eps=1e-12): 
    r""" Main function to register two input volumes using Demon's registration algorithm in the SimpleITK library in a multiscale manner. 
    
    see SimpleITK notebooks, https://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/Python_html/66_Registration_Demons.html

    Parameters
    ----------
    vol1 : (MxNxL) array
        reference volume given as a Numpy array
    vol2 : (MxNxL) array
        volume to register to vol2, given as a Numpy array
    imtype : int
        specifies the -bit of the image e.g. 8-bit, 16-bit etc. Used to normalize the intensity if ``rescale_intensity=False``
    p12 : (p1,p2) tuple
        if provided, vol1 and vol2 are constract enhanced using percentage intensity normalization 
    rescale_intensity : bool
        if True, does min-max scaling of image intensities in the two volumes independently before registering. This follows percentage intensity normalization if specified. 
    centre_tfm_model : str
        specifies the initial centering model. This is one of two options. For microscopy, we find 'geometry' has less jitter and is best. 
        
        'geometry': str
            uses sitk.CenteredTransformInitializerFilter.GEOMETRY. This computes the geometrical center independent of intensity.
        'moments': str
            uses sitk.CenteredTransformInitializerFilter.MOMENTS. This computes the geometrical center of mass of the volume based on using image intensity as a weighting. 

    demons_type : str
        One of two demon's filters in SimpleITK. 

        'diffeomorphic' : str
            sitk.DiffeomorphicDemonsRegistrationFilter(). This implements the diffeomorphic demons approach of [1]_ to penalise foldovers in the warp field and is more biologically plausible
        'symmetric' : str
            sitk.FastSymmetricForcesDemonsRegistrationFilter(). This implements the symmetric forces of [2]_. The idea is that in general the warp field of registering vol2 to vol1 and vice versa is not quite the same. in Symmetric forces demon the learning takes the average of these two warp fields to ensure symmetry preservation.

    shrink_factors : list of lists or scalars 
        Shrink factors relative to the original image's size. When the list entry, shrink_factors[i], is a scalar the same factor is applied to all axes.
        When the list entry is a list, shrink_factors[i][j] is applied to axis j. This allows us to specify different shrink factors per axis. This is useful
        in the context of microscopy images where it is not uncommon to have unbalanced sampling such as a 512x512x8 image. In this case we would only want to 
        sample in the x,y axes and leave the z axis as is: [[[8,8,1],[4,4,1],[2,2,1]].
    smoothing_sigmas : list of lists or scalars  
        Amount of smoothing which is done prior to resmapling the image using the given shrink factor. These are in pixel units.
    eps : scalar
        small number for numerical precision
    
    Returns
    -------
    SimpleITK.DisplacementFieldTransform

    References
    ----------
    .. [1] Vercauteren et al. "Diffeomorphic demons: Efficient non-parametric image registration." NeuroImage 45.1 (2009): S61-S72.
    .. [2] Avants et al. "Symmetric diffeomorphic image registration with cross-correlation: evaluating automated labeling of elderly and neurodegenerative brain." Medical image analysis 12.1 (2008): 26-41.

    """
    import SimpleITK as sitk

    def imadjust(vol, p1, p2): 
        import numpy as np 
        from skimage.exposure import rescale_intensity
        # this is based on contrast stretching and is used by many of the biological image processing algorithms.
        p1_, p2_ = np.percentile(vol, (p1,p2))
        vol_rescale = rescale_intensity(vol, in_range=(p1_,p2_))
        return vol_rescale

    import numpy as np 

    if p12 is None:
        im1_ = vol1.copy()
        im2_ = vol2.copy()
        # no contrast stretching 
        if rescale_intensity == False:
            im1_ = im1_ / (2**imtype - 1)
            im2_ = im2_ / (2**imtype - 1)
        else:
            im1_ = (im1_ - im1_.min()) / (im1_.max() - im1_.min() + eps)
            im2_ = (im2_ - im2_.min()) / (im2_.max() - im2_.min() + eps)
    else:
        # contrast stretching 
        im1_ = imadjust(vol1, p12[0], p12[1])
        im2_ = imadjust(vol2, p12[0], p12[1])

        if rescale_intensity == False:
            im1_ = im1_ / (2**imtype - 1)
            im2_ = im2_ / (2**imtype - 1)
        else:
            im1_ = (im1_ - im1_.min()) / (im1_.max() - im1_.min() + eps)
            im2_ = (im2_ - im2_.min()) / (im2_.max() - im2_.min() + eps)

    ### main analysis scripts. 
    v1 = sitk.GetImageFromArray(im1_, isVector=False)
    v2 = sitk.GetImageFromArray(im2_, isVector=False)
        
    # a) initial transform 
    # translation.
    if centre_tfm_model=='geometry':
        translation_mode = sitk.CenteredTransformInitializerFilter.GEOMETRY
    if centre_tfm_model=='moments':
        translation_mode = sitk.CenteredTransformInitializerFilter.MOMENTS
    initial_transform = sitk.CenteredTransformInitializer(v1, 
                                                          v2, 
                                                          sitk.Euler3DTransform(),
                                                          translation_mode)

    # a) demons transform (best to have corrected out any rigid transforms a priori) 
     # Select a Demons filter and configure it.
    if demons_type == 'diffeomorphic': 
        demons_filter = sitk.DiffeomorphicDemonsRegistrationFilter() # we should use this version.
    if demons_type == 'symmetric':
        demons_filter = sitk.FastSymmetricForcesDemonsRegistrationFilter()
    # set the number of iterations
    demons_filter.SetNumberOfIterations(n_iters) # 5 for less. # long time for 20? 
    # Regularization (update field - viscous, total field - elastic).
    demons_filter.SetSmoothDisplacementField(smooth_displacement_field)
    demons_filter.SetStandardDeviations(smooth_alpha)
    
    # run the registration and return the final transform parameters
    final_tfm = multiscale_demons(registration_algorithm=demons_filter, 
                                  fixed_image = v1,
                                  moving_image = v2,
                                  initial_transform = initial_transform,
                                  shrink_factors = shrink_factors, # did have 2 here. -> test, can we separate the  # do at the same scale. 
                                  smoothing_sigmas = smoothing_sigmas) # set smoothing very low, since we want it to zone in on interesting features. 
    # check again how this is parsed .
    return final_tfm

# define helper scripts to transform new volumes given a deformation.
def transform_img_sitk(vol, tfm):
    r""" One-stop function for applying any SimpleITK transform to an input image. Linear interpolation is used.  
    
    Parameters
    ----------
    vol: array
        input image as a numpy array     
    tfm: SimpleITK.Transform
        A simpleITK transform instance such as that resulting from using the simpleITK registration functions in this module.
    
    Returns
    -------
    v_transformed : array
        resulting image after applying the given transform to the input, returned as a numpy array 

    """
    import SimpleITK as sitk

    v = sitk.GetImageFromArray(vol, isVector=False)
    v_transformed = sitk.Resample(v, 
                                  v, 
                                  tfm, # this should work with all types of transforms.
                                  sitk.sitkLinear, 
                                  0.0, 
                                  v.GetPixelID())
    v_transformed = sitk.GetArrayFromImage(v_transformed) # back to numpy format. 
    
    return v_transformed


# here we need to add functions for reading and writing SITK transforms, and converting a numpy array to diplacemeent field to take into advantage of SITK image resampling. 
